#!/usr/bin/env python3

"""Tests asciidoc3-generated xml (or of course any other xml) files
   against the docbook51 schematron schema - heuristic approach.

   USAGE:
   'python3 asciidoc3_docbook51_schematron_validation.py'
   to test all files ending with 'docbook51.xml' in directory './tests/data' or
   'python3 asciidoc3_docbook51_schematron_validation.py path/to/db51_xml_file.xml'
   to validate just one file.
   Example:
   'python3 asciidoc3.py -b docbook51 ./tests/data/nonvalid_docbook51.txt'
   and now
   'python3 asciidoc3_docbook51_schematron_validation.py ../data/nonvalid_docbook51.xml'
   to see the test finding a non-valid document (which relaxng doesn't detect.
   
   This is a BETA version, provided "as is", no warranty.
   Please note: False positive and false negative results are possible.

   (c) 2020 by Berthold Gehrke <berthold.gehrke@gmail.com>
   License: GNU General Public License v2 or higher (GPLv2+)
   """

import xml.etree.ElementTree as et
from io import StringIO
import os
from pathlib import Path
import re
import sys


def schematron51_check(db51testfile):
    # print("!beta! Starting schematron validation DocBook5.1: ", db51testfile)
    try:
        with open(path_to_data + db51testfile, 'r') as testfile:
            testcontent = testfile.read()
    except (MemoryError, OSError):
        sys.exit("Giving up - File to large?", db51testfile)
    testcontent, counter = re.subn(r'\A.*?(<\?xml.*?>)\s*?<',
                                   '<', testcontent, 1, flags=re.DOTALL)
    testcontent, counter = re.subn(r'\A<\?.*?toc\?>\s*<',
                                   '<', testcontent, 1, flags=re.DOTALL)
    testcontent, counter = re.subn(r'\A<\?.*?numbered\?>\s*<',
                                   '<', testcontent, 1, flags=re.DOTALL)
    # root element must have an attribute "version"
    testcontent, counter = re.subn(r'\A(<[^>]*?version=[^>]*?>\s*<)',
                                   '\g<1>', testcontent, 1, flags=re.DOTALL)
    if not counter:
        print("*!*!*!*!", db51testfile, "not valid!")
        print(" Must have an attribute 'version' in root element: not found.")

    try:
        testcontent = StringIO(testcontent)
        tree = et.parse(testcontent)
        t = tree.getroot()
    except (MemoryError, OSError):
        sys.exit("Giving up - File to large?", db51testfile)
    # At this time we have "t" the etree of file "db51testfile" in memory.
    # We can delete "testcontent" to spare some memory.
    del(testcontent)
    # Now we fill the lists of elements which have restrictions in
    # DocBook51 Schematron.
    
    # 'seglistitem' not used in Asciidoc3
    # 'annotation' not used in Asciidoc3
    caution_list = list()
    important_list = list()
    note_list = list()
    tip_list = list()
    warning_list = list()
    caption_list = list()
    equation_list = list()
    example_list = list()
    figure_list = list()
    table_list = list()
    footnote_list = list()
    sidebar_list = list()
    entry_list = list()
    footnoteref_list = list()
    # 'firstterm' not used in Asciidoc3
    glossterm_list = list()  # TODO
    # 'glossseealso' not used in Asciidoc3
    # 'termdef' not used in Asciidoc3
    # 'glosssee' not used in Asciidoc3
    indexterm_list = list()  # TODO
    # 'synopfragmentref' not used in Asciidoc3
    # 'xlink:type 'arc', 'extended', 'locator', 'resource',
    # and 'title' are not used in Asciidoc3.

    # Iter through the entire tree to fill the lists.
    for elem in t.iter(tag='*'):
        if elem.tag == '{http://docbook.org/ns/docbook}caution':
            caution_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}important':
            important_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}note':
            note_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}tip':
            tip_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}warning':
            warning_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}caption':
            caption_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}equation':
            equation_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}example':
            example_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}figure':
            figure_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}table':
            table_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}footnote':
            footnote_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}sidebar':
            sidebar_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}entry':
            entry_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}footnoteref':
            footnoteref_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}glossterm':
            glossterm_list.append(elem)
        elif elem.tag == '{http://docbook.org/ns/docbook}indexterm':
            indexterm_list.append(elem)
    # Now we define the specific 'error function' regarding the elements.

    def caution_error():
        caution_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in caution_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in caution_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'caution'")

    def important_error():
        important_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in important_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in important_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'important'")

    def note_error():
        note_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in note_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in note_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'note'")

    def tip_error():
        tip_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in tip_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in tip_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'tip'")

    def warning_error():
        warning_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in warning_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in warning_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'warning'")

    def caption_error():
        caption_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure',
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}sidebar',
            '{http://docbook.org/ns/docbook}table',
            '{http://docbook.org/ns/docbook}task',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in caption_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in caption_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'caption'")

    def equation_error():
        equation_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure',
            '{http://docbook.org/ns/docbook}table', ])
        for node in equation_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in equation_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'equation'")

    def example_error():
        example_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure',
            '{http://docbook.org/ns/docbook}table', ])
        for node in example_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in example_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'example'")

    def figure_error():
        figure_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure',
            '{http://docbook.org/ns/docbook}table', ])
        for node in figure_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in figure_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'figure'")

    def table_error():
        table_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure', ])
        for node in table_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in table_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'table'")

    def footnote_error():
        footnote_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}caution',
            '{http://docbook.org/ns/docbook}epigraph',
            '{http://docbook.org/ns/docbook}equation',
            '{http://docbook.org/ns/docbook}example',
            '{http://docbook.org/ns/docbook}figure',
            '{http://docbook.org/ns/docbook}footnote',
            '{http://docbook.org/ns/docbook}important',
            '{http://docbook.org/ns/docbook}note',
            '{http://docbook.org/ns/docbook}sidebar',
            '{http://docbook.org/ns/docbook}table',
            '{http://docbook.org/ns/docbook}task',
            '{http://docbook.org/ns/docbook}tip',
            '{http://docbook.org/ns/docbook}warning', ])
        for node in footnote_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in footnote_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'footnote'")

    def sidebar_error():
        sidebar_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}sidebar', ])
        for node in sidebar_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in sidebar_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'sidebar'")

    def entry_error():
        entry_not_allowed = frozenset([
            '{http://docbook.org/ns/docbook}informaltable',
            '{http://docbook.org/ns/docbook}table', ])
        for node in entry_list:
            for item in list(node.iter('*'))[1:]:
                if item.tag in entry_not_allowed:
                    print("*!*!*!*!", db51testfile, "not valid!")
                    print(" Found " + item.tag[31:] + " as")
                    print(" descendant of 'entry'")

    def footnoteref_error():
        """ beta, TODO """
        footnoteref_linkend_list = list()
        footnote_anchor_list = list()
        for node in footnoteref_list:
            footnoteref_linkend_list.append(node.get('linkend'))
            # print("footnoteref_linkend_list ", footnoteref_linkend_list)
            # There is at least one "footnoteref linkend", but not a "footnote" at all
            if footnoteref_linkend_list and not footnote_list:
                print("*!*!*!*!", db51testfile, "not valid!")
                print(" Did not found any footnote,")
                print(" but there is a 'footnoteref linkend'")
            # document contains both "footnoteref linkend" and "footnote"
            elif footnoteref_linkend_list and footnote_list:
                for item in footnote_list:
                    footnote_anchor_list.append(item.get(
                        '{http://www.w3.org/XML/1998/namespace}id'))
                # print("footnote_anchor_list", footnote_anchor_list)
                for item_fl in footnoteref_linkend_list:
                    if item_fl not in footnote_anchor_list:
                        print("*!*!*!*!", db51testfile, "not valid!")
                        print(" No '" + item_fl + "' as ref in any footnote")
            else:
                pass  # valid so far (document contains no "footnoteref linkend")
                
    def glossterm_error():
        """ TODO """
        pass

    def indexterm_error():
        """ TODO """
        pass

    if caution_list:
        caution_error()
    if important_list:
        important_error()
    if note_list:
        note_error()
    if tip_list:
        tip_error()
    if warning_list:
        warning_error()
    if caption_list:
        caption_error()
    if equation_list:
        equation_error()
    if example_list:
        example_error()
    if figure_list:
        figure_error()
    if table_list:
        table_error()
    if footnote_list:
        footnote_error()
    if sidebar_list:
        sidebar_error()
    if entry_list:
        entry_error()
    if footnoteref_list:
        footnoteref_error()
    if glossterm_list:
        glossterm_error()
    if indexterm_list:
        indexterm_error()


def main():
    if filepath:
        schematron51_check(base_filepath)
    else:
        with os.scandir(path_to_data) as testpath:
            for entry in testpath:
                if entry.name.endswith('docbook51.xml'):
                    schematron51_check(entry.name)


if __name__ == '__main__':
    filepath = ''
    full_path = os.path.abspath(os.path.dirname(__file__))
    if len(sys.argv) == 2:
        filepath = sys.argv[1]
        if filepath.endswith('/'):
            sys.exit("FATAL: Please enter path without trailing '/'!")
        base_filepath = os.path.basename(filepath)
        dirname = os.path.dirname(filepath)
        if dirname:
            path_to_data = dirname + "/"
        else:
            path_to_data = ""
    elif len(sys.argv) > 2:
        sys.exit("FATAL: Only one argument (path) possible!")
    elif len(sys.argv) == 1:
        path_to_tests = str(Path(full_path).parents[0])
        path_to_data = path_to_tests + "/data/"
    else:
        sys.exit("FATAL: Found no argument at all!")
    main()



#!/usr/bin/env python3

"""
Runs a bunch of a2x3.py testcases analogue to 'testasciidoc3.py'
(which tests asciidoc3.py). First some files are copied to
./tests/data/a2x3testdata
In a second step a2x3.py is executed in diverse approaches. See the
'command lines' presented as 'TEST_TUPEL'.
Be sure to have all needed programs installed: lynx, w3m, epubcheck,
dblatex, fop ...

Before you can use testa2x3.py you have to run
'python3 testasciidoc3.py --force update'
to produce the testfiles. Output is written to
'./tests/data/a2x3testdata/'

Usage: [python3 testasciidoc3.py --force update &&] python3 testa2x3.py [-v]

Known bugs:
If producing dvi/ps, the source must not contain images - this seems to
be a dblatex issue ... (?)

Copyright (C) 2020 by Berthold Gehrke <berthold.gehrke@gmail.com>
Free use of this software is granted under the terms of the
GNU General Public License Version 2 or higher (GNU GPLv2+).
"""

from argparse import ArgumentParser
from concurrent import futures
import os
import shutil
import subprocess


TEST_TUPEL = (
    # txt to pdf
    ['-f', 'pdf', './data/a2x3testdata/pdf_test.txt'],
    # txt to pdf using fop
    ['-f', 'pdf', '--fop', './data/a2x3testdata/fop_test.txt'],
    # txt to epub
    ['-f', 'epub', './data/a2x3testdata/epub_test.txt'],
    # txt to epub leave artifacts
    ['-f', 'epub', '-k', './data/a2x3testdata/epub_artifacts_test.txt'],
    # txt to epub epubcheck
#    ['-f', 'epub', '-k', '--epubcheck', \
#     './data/a2x3testdata/epubcheck_test.txt'],
    # txt to text (lynx)
    ['-f', 'text', '--lynx', './data/a2x3testdata/lynx_test.txt'],
    # txt to text (w3m)
    ['-f', 'text', './data/a2x3testdata/w3m_test.txt'],
    # txt to xhtml
    ['-f', 'xhtml', './data/a2x3testdata/test.txt'],
    # txt to manpage
    ['-f', 'manpage', '-d', 'manpage', './data/a2x3testdata/a2x3.1.txt'],
    # txt to docbook xml
    ['-f', 'docbook', './data/a2x3testdata/docbook_test.txt'],
    # xml (docbook) to pdf
    ['-f', 'pdf', './data/a2x3testdata/test-docbook45.xml'],
    # txt to chunked
    ['-f', 'chunked', './data/a2x3testdata/chunked_test.txt'],
    # xml (docbook) to dvi
    ['-f', 'dvi', './data/a2x3testdata/faq_dvi_test_45.xml'],
    # xml (docbook) to ps
    ['-f', 'ps', './data/a2x3testdata/faq_ps_test_51.xml'],
    # xml (docbook) to tex
    ['-f', 'tex', './data/a2x3testdata/tex_test_51.xml'],
    )


DESCRIPTION = '''Runs tests for AsciiDoc3 / a2x3.py.
                 Use option '-v' to see a2x3 working.'''
ARG_PARSER = ArgumentParser(usage='usage: python3 a2x3test.py [-v]',
                            description=DESCRIPTION)
ARG_PARSER.add_argument('-v', '--verbose', action='store_true', default=False,
                        help='see more output')
ARGS = ARG_PARSER.parse_args()


def remove_testitems():
    """remove files from previous tests"""
    if os.path.exists('./data/a2x3testdata/'):
        shutil.rmtree('./data/a2x3testdata/')
        if ARGS.verbose:
            print("remove ./data/a2x3testdata/")


def copy_testfiles():
    """Copy files to ./data/a2x3testdata/"""
    os.mkdir('./data/a2x3testdata/')
    for file_item in ('test.txt',
                      'a2x3.1.txt',
                      ):
        if os.path.exists('../doc/'+file_item):
            shutil.copy2('../doc/'+file_item, './data/a2x3testdata/'+file_item)
            if ARGS.verbose:
                print("copy "+file_item)
    # We use copies of 'test.txt' to test text, pdf, fop, epub, epubcheck ...'
    if os.path.exists('./data/a2x3testdata/test.txt'):
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/pdf_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/fop_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/lynx_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/w3m_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/fop_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/epub_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/epub_artifacts_test.txt')
#        shutil.copy2('./data/a2x3testdata/test.txt',
#                     './data/a2x3testdata/epubcheck_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/docbook_test.txt')
        shutil.copy2('./data/a2x3testdata/test.txt',
                     './data/a2x3testdata/chunked_test.txt')
        shutil.copy2('./data/test-docbook45.xml',
                     './data/a2x3testdata/test-docbook45.xml')
        shutil.copy2('./data/faq-docbook45.xml',
                     './data/a2x3testdata/faq_dvi_test_45.xml')
        shutil.copy2('./data/faq-docbook51.xml',
                     './data/a2x3testdata/faq_ps_test_51.xml')
        shutil.copy2('./data/test-docbook51.xml',
                     './data/a2x3testdata/tex_test_51.xml')
        if ARGS.verbose:
            print("copy testfiles to ./data/a2x3testdata/")


def run_tests_a2x3(list_data):
    """Runs tests parallel"""
    arg_list = ['python3', '../a2x3.py']
    if ARGS.verbose:
        arg_list += ['-v']
    arg_list += list_data
    if ARGS.verbose:
        print(arg_list)
    subprocess_a2x3test = subprocess.run(arg_list)
    if subprocess_a2x3test.returncode:
        print("[WARN] Returncode != zero running", arg_list)


if __name__ == '__main__':
    remove_testitems()
    copy_testfiles()
    with futures.ProcessPoolExecutor(max_workers=4) as e:
        for test_item in TEST_TUPEL:
            e.submit(run_tests_a2x3, test_item)



#!/usr/bin/env python3

'''
a2x3 - A toolchain manager for AsciiDoc3 - converts AsciiDoc3 text
files to other file formats. See asciidoc3.org for more information.

Copyright: (c) 2009 Stuart Rackham <srackham@gmail.com> MIT
Copyright: (c) 2018-2020 Berthold Gehrke <berthold.gehrke@gmail.com>
               for Python3 version
License:   GNU GPL v2 or higher
'''

from argparse import ArgumentParser
import fnmatch
import html.parser
import mimetypes
import os
import re
import shutil
import subprocess
import sys
import traceback
import urllib.parse
import xml.dom.minidom
import zipfile


PROG = os.path.basename(os.path.splitext(__file__)[0])
# Version corresponds to asciidoc3.py version
VERSION = '3.2.0'

# AsciiDoc3 global configuration file directory.
# NOTE: CONF_DIR is "fixed up" during install -- don't rename or change syntax.
CONF_DIR = '/etc/asciidoc3'
# if installed 'local', e.g. only in your home dir, you may use your directory
# here; but works somehow or other ...
# CONF_DIR = '~/.asciidoc3'
# Update 2019-May: there is no need for action anyway ...

# ####################################################################
# Default configuration file parameters.
# ####################################################################

# Optional environment variable dictionary passed to executing programs.
# If set to None the existing environment is used.
ENV = None

# External executables.
ASCIIDOC3 = 'asciidoc3'
XSLTPROC = 'xsltproc'
DBLATEX = 'dblatex'         # pdf generation.
FOP = 'fop'                 # pdf generation (--fop option).
W3M = 'w3m'                 # text generation.
LYNX = 'lynx'               # text generation (if no w3m).
XMLLINT = 'xmllint'         # Set to '' to disable.
# XMLLINT = ''                # Do not use xmllint (= skip xml validation)
EPUBCHECK = 'epubcheck'     # Set to '' to disable.
# External executable default options.
ASCIIDOC3_OPTS = ''
DBLATEX_OPTS = ''
FOP_OPTS = ''
XSLTPROC_OPTS = ''
BACKEND_OPTS = ''
# Location of RelaxNG Schema for DocBook5.1 to validate the xml-file.
# Full path needed, example
# '/home/myname/asciidoc3/docbook_v51/rng-docbook.rng'.
# In the very most cases you may leave this empty, and a2x3.py
# finds the file at the default place in the AsciiDoc3 distribution.
relaxng_location = ''

# ####################################################################
# End of configuration file parameters.
# ####################################################################


# ####################################################################
# Utility functions
# ####################################################################

OPTIONS = None  # These functions read verbose and dry_run command options.


def errmsg(msg):
    """ Prints error message and underlying program to stderr."""
    sys.stderr.write('%s: %s\n' % (PROG, msg))


def warning(msg):
    """ Prints warning message to stdout."""
    errmsg('WARNING: %s' % msg)


def infomsg(msg):
    """ Prints info message and underlying program to stdout."""
    print('%s: %s' % (PROG, msg))


def die(msg, exit_code=1):
    """ Prints fatal message to stdout and exits."""
    errmsg('ERROR: %s' % msg)
    sys.exit(exit_code)


def trace():
    """Print traceback to stderr."""
    errmsg('-'*60)
    traceback.print_exc(file=sys.stderr)
    errmsg('-'*60)


def verbose(msg):
    """ Prints verbose message to stdout."""
    if OPTIONS.verbose or OPTIONS.dry_run:
        infomsg(msg)


class AttrDict(dict):
    """
    Like a dictionary except values can be accessed as attributes i.e. obj.foo
    can be used in addition to obj['foo'].
    If self._default has been set then it will be returned if a non-existent
    attribute is accessed (instead of raising an AttributeError).
    """
    def __getattr__(self, key):
        try:
            return self[key]
        except KeyError as k:
            if '_default' in self:
                return self['_default']
            raise AttributeError(k)

    def __setattr__(self, key, value):
        self[key] = value

    def __delattr__(self, key):
        try:
            del self[key]
        except KeyError as k:
            raise AttributeError(k)

    def __repr__(self):
        return '<AttrDict ' + dict.__repr__(self) + '>'

    def __getstate__(self):
        return dict(self)

    def __setstate__(self, value):
        for item_key, item_value in list(value.items()):
            self[item_key] = item_value


def isexecutable(file_name):
    """ Returns True if file_name is executable -
        surprise, just as the function is named ..."""
    return os.path.isfile(file_name) and os.access(file_name, os.X_OK)


def find_executable(file_name):
    '''
    Search for executable file_name in the system PATH.
    Return full path name or None if not found.
    '''
    def _find_executable(file_name):
        if os.path.split(file_name)[0] != '':
            # file_name includes directory so don't search path.
            if not isexecutable(file_name):
                return None
            return file_name
        for path_item in os.environ.get('PATH', os.defpath).split(os.pathsep):
            file_candidate = os.path.join(path_item, file_name)
            if isexecutable(file_candidate):
                return os.path.realpath(file_candidate)
        return None
    if os.name == 'nt' and os.path.splitext(file_name)[1] == '':
        for ext in ('.cmd', '.bat', '.exe'):
            result = _find_executable(file_name + ext)
            if result:
                break
    else:
        result = _find_executable(file_name)
    return result


def write_file(filename, data, mode='w'):
    """ Writes 'data' to 'filename' """
    with open(filename, mode) as file_to_write:
        file_to_write.write(data)


def read_file(filename, mode='r'):
    """ Reads from 'filename' """
    with open(filename, mode) as file_to_read:
        return file_to_read.read()


def shell_cd(path):
    """ change to directory 'path' """
    verbose('chdir %s' % path)
    if not OPTIONS.dry_run:
        os.chdir(path)


def shell_makedirs(path):
    """ Creates directories 'path' """
    if os.path.isdir(path):
        return
    verbose('creating %s' % path)
    if not OPTIONS.dry_run:
        os.makedirs(path)


def shell_copy(src, dst):
    """ Copy files src to dst """
    verbose('copying "%s" to "%s"' % (src, dst))
    if not OPTIONS.dry_run:
        shutil.copy(src, dst)


def shell_rm(path):
    """ Deletes a file 'path' """
    if not os.path.exists(path):
        return
    verbose('deleting %s' % path)
    if not OPTIONS.dry_run:
        os.unlink(path)


def shell_rmtree(path):
    """ Deletes directories 'path' """
    if not os.path.isdir(path):
        return
    verbose('deleting %s' % path)
    if not OPTIONS.dry_run:
        shutil.rmtree(path)


def shell(cmd, raise_error=True):
    '''
    Execute command cmd in shell and return tuple
    (stdoutdata, stderrdata, returncode).
    If raise_error is True then a non-zero return terminates the application.
    '''
    if os.name == 'nt':
        # This is probably unnecessary, see:
        # http://groups.google.com/group/asciidoc/browse_frm/thread/9442ee0c419f1242
        # Windows doesn't like running scripts directly so explicitly
        # specify python interpreter.
        # Extract first (quoted or unquoted) argument.
        match_object = re.match(r'^\s*"\s*(?P<arg0>[^"]+)\s*"', cmd)
        if not match_object:
            match_object = re.match(r'^\s*(?P<arg0>[^ ]+)', cmd)
        if match_object.group('arg0').endswith('.py'):
            cmd = 'python3 ' + cmd
        # Remove redundant quoting -- this is not just cosmetic,
        # quoting seems to dramatically decrease the allowed command
        # length in Windows XP.
        cmd = re.sub(r'"([^ ]+?)"', r'\1', cmd)
    verbose('executing: %s' % cmd)
    if not OPTIONS.dry_run:
        stdout = stderr = subprocess.PIPE
        try:
            # popen(...) was changed to fit Python3's string/utf-8 definitions
            popen = subprocess.Popen(cmd, stdout=stdout, stderr=stderr,
                                     shell=True, env=ENV,
                                     universal_newlines=True, bufsize=-1)
        except OSError as os_error:
            die('failed: %s: %s' % (cmd, os_error))
        stdoutdata, stderrdata = popen.communicate()
        if OPTIONS.verbose:
            print(stdoutdata)
            print(stderrdata)
        if popen.returncode != 0 and raise_error:
            die('%s returned non-zero exit status %d'
                % (cmd, popen.returncode))
        return (stdoutdata, stderrdata, popen.returncode)


def find_resources(files, tagname, attrname, _filter=None):
    '''
    Search all files and return a list of local URIs from attrname attribute
    values in tagname tags.
    Handles HTML open and XHTML closed tags.
    Non-local URIs are skipped.
    files can be a file name or a list of file names.
    The _filter function takes a dictionary of tag attributes and returns
    True if the URI is to be included.

    Added in a2x3.py / AsciiDoc3:
    A call of 'find_resources' is found in
    'class A2X3 def copy_resources(...)':
    ... find_resources(html_files, 'link', 'href',
                       lambda attrs: attrs.get('type') == 'text/css') ...
    That means: parse file 'foo.html' (that is one of or just the
    'html_files') to find tags '<link ...  >' - '<a class="link" href="#X40">'
    is not found. A sample is
    '<link rel="stylesheet" type="text/css" href="docbook-xsl.css" />'
    because 'type' == 'text/css' and "_filter(attrs)" == 'True'.
    So the content of attribute 'href' (== docbook-xsl.css) is a
    resource-candidate. 'docbook-xsl.css' will be appended to 'result'
    if 'uri' is a 'local file'.
    A second call is
    ... find_resources(html_files, 'img', 'src') ...
    That means: parse file 'foo.html' (that is one of or just the 'html_files')
    to find tags '<img ...  >'. A sample is
    '<img alt="[Tip]" src="images/icons/tip.png" />' because the tag starts
    with '<img'. So the content of attribute 'src' (== images/icons/tip.png)
    is another resource-candidate. 'images/icons/tip.png' will be appended
    to 'result' if 'uri' is a 'local file'.
    '''
    class FindResources(html.parser.HTMLParser):
        """ Nested parser class shares locals with enclosing function."""

        def handle_startendtag(self, tag, attrs):
            """ startendtags, e.g. '<img ... />', are handled identically
                like starttags, e.g. '<img ... >' """
            self.handle_starttag(tag, attrs)

        def handle_starttag(self, tag, attrs):
            attrs = dict(attrs)
            if tag == tagname and (_filter is None or _filter(attrs)):
                # Accept only local URIs.
                uri = urllib.parse.urlparse(attrs[attrname])
                if uri[0] in ('', 'file') and not uri[1] and uri[2]:
                    result.append(uri[2])

        def error(self, message):
            """ See ... cpython-master/Lib/_markupbase.py. Unused;
                to avoid pylint3 message: W0223: Method 'error' is abstract
                in class 'ParserBase' but is not overridden"""
            try:
                verbose("""Error in 'class FindResources/def
                           find_resources()': {}. Does nothing - we keep
                           our fingers crossed. """.format(message))
            except OSError:
                verbose("""Something went wrong (OSError) in
                        'class FindResources def find_resources()':
                        We keep our fingers crossed.""")

    if isinstance(files, str):
        files = [files]
    result = []
    for filename in files:
        verbose('finding resources in: %s' % filename)
        if OPTIONS.dry_run:
            continue
        parser = FindResources()
        # HTMLParser has problems with non-ASCII strings.
        # See http://bugs.python.org/issue3932
        # a2x3: opens all html-files in 'utf-8' and ignores errors.
        # The encoding of the 'data' (the text not included by tags)
        # doesn't care - it is unused information. We need only the
        # information in tag-attributes and may assume 'utf-8' ...
        contents = read_file(filename)
        match_object = re.search(r'\A<\?xml.* encoding="(.*?)"', contents)
        if match_object:
            # encoding = match_object.group(1)  # v2
            # _encoding = match_object.group(1) # v3 (unused var _encoding)
            # parser.feed(contents.decode(encoding))                   # v2
            contents = str(contents.encode('utf-8', errors='ignore'))  # v3
        parser.feed(contents)
        parser.close()
    result = sorted(set(result))   # Drop duplicate values.
    return result


def find_files(path, pattern):
    '''
    Return list of file names matching pattern in directory path.
    '''
    result = []
    for (file_path, __, files) in os.walk(path):
        for file_item in files:
            if fnmatch.fnmatch(file_item, pattern):
                result.append(
                    os.path.normpath(os.path.join(file_path, file_item)))
    return result


def exec_xsltproc(xsl_file, xml_file, dst_dir, opts=''):
    """ Executes appropriate xslt cmd."""
    cwd = os.getcwd()
    shell_cd(dst_dir)
    try:
        shell('"%s" %s "%s" "%s"' % (XSLTPROC, opts, xsl_file, xml_file))
    finally:
        shell_cd(cwd)


def get_source_options(asciidoc3_file):
    '''
    Look for a2x3 command options in AsciiDoc3 source file.
    Limitation: options cannot contain double-quote characters.
    '''
    def parse_options():
        # Parse options to result sequence.
        inquotes = False
        opt = ''
        for option_item in options:
            if option_item == '"':
                if inquotes:
                    result.append(opt)
                    opt = ''
                    inquotes = False
                else:
                    inquotes = True
            elif option_item == ' ':
                if inquotes:
                    opt += option_item
                elif opt:
                    result.append(opt)
                    opt = ''
            else:
                opt += option_item
        if opt:
            result.append(opt)

    result = []
    if os.path.isfile(asciidoc3_file):
        options = ''
        with open(asciidoc3_file) as ad3_file:
            for line in ad3_file:
                match_object = re.search(r'^//\s*a2x3:', line)
                if match_object:
                    options += ' ' + line[match_object.end():].strip()
        parse_options()
    return result

# ###################################################################
# Application class
# ###################################################################


class A2X3(AttrDict):
    '''
    a2x3 options and conversion functions.
    '''

    def execute(self):
        '''
        Process a2x3 command.
        '''
        self.process_options()
        # Append configuration file options.
        self.asciidoc3_opts += ' ' + ASCIIDOC3_OPTS
        self.dblatex_opts += ' ' + DBLATEX_OPTS
        self.fop_opts += ' ' + FOP_OPTS
        self.xsltproc_opts += ' ' + XSLTPROC_OPTS
        self.backend_opts += ' ' + BACKEND_OPTS
        # Execute to_* functions.
        if self.backend:
            self.to_backend()
        else:
            self.__getattribute__('to_'+self.format)()
        if not (self.keep_artifacts or self.format == 'docbook'
                or self.skip_asciidoc3):
            shell_rm(self.dst_path('.xml'))

    def load_conf(self):
        '''
        Load a2x3 configuration file from default locations and --conf-file
        option.
        '''
        # global ASCIIDOC3
        primary_conf_file = 'a2x3.conf'
        a2x3dir = os.path.dirname(os.path.realpath(__file__))
        conf_files = []
        # From a2x3.py directory.
        conf_files.append(os.path.join(a2x3dir, primary_conf_file))
        # If the asciidoc3 executable and conf files are in the a2x3 directory
        # then use the local copy of asciidoc3 and skip the global a2x3 conf.
        asciidoc3 = os.path.join(a2x3dir, 'asciidoc3.py')
        asciidoc3_conf = os.path.join(a2x3dir, 'asciidoc3.conf')
        if os.path.isfile(asciidoc3) and os.path.isfile(asciidoc3_conf):
            self.asciidoc3 = asciidoc3
        else:
            self.asciidoc3 = None
            # From global conf directory.
            conf_files.append(os.path.join(CONF_DIR, primary_conf_file))
        # From $HOME directory.
        home_dir = os.environ.get('HOME')
        if home_dir is not None:
            conf_files.append(os.path.join(home_dir,
                                           '.asciidoc3', primary_conf_file))
        # If asciidoc3 is not local to a2x3 then search the PATH.
        if not self.asciidoc3:
            self.asciidoc3 = find_executable(ASCIIDOC3)
            if not self.asciidoc3:
                die('unable to find asciidoc3: %s' % ASCIIDOC3)
        # From backend plugin directory.
        if self.backend is not None:
            stdout = shell(self.asciidoc3 + ' --backend list')[0]
            backends = [(i, os.path.split(i)[1]) for i in stdout.splitlines()]
            backend_dir = [i[0] for i in backends if i[1] == self.backend]
            if not backend_dir:
                die('missing %s backend' % self.backend)
            if len(backend_dir) > 1:
                die('more than one %s backend' % self.backend)
            verbose('found %s backend directory: %s' %
                    (self.backend, backend_dir[0]))
            conf_files.append(os.path.join(backend_dir[0], 'a2x3-backend.py'))
        # From --conf-file option.
        if self.conf_file is not None:
            if not os.path.isfile(self.conf_file):
                die('missing configuration file: %s' % self.conf_file)
            conf_files.append(self.conf_file)
        # From --xsl-file option.
        if self.xsl_file is not None:
            if not os.path.isfile(self.xsl_file):
                die('missing XSL file: %s' % self.xsl_file)
            self.xsl_file = os.path.abspath(self.xsl_file)
        # Load ordered files.
        for file_item in conf_files:
            if os.path.isfile(file_item):
                verbose('loading configuration file: %s' % file_item)
                exec(compile(open(file_item).read(),
                             file_item, 'exec'), globals())

    def process_options(self):
        '''
        Validate and command options and set defaults.
        '''
        if not os.path.isfile(self.asciidoc3_file):
            die('missing SOURCE_FILE: %s' % self.asciidoc3_file)
        self.asciidoc3_file = os.path.abspath(self.asciidoc3_file)
        if os.path.splitext(self.asciidoc3_file)[1].lower() == '.xml':
            self.skip_asciidoc3 = True
        else:
            self.skip_asciidoc3 = False
        if not self.destination_dir:
            self.destination_dir = os.path.dirname(self.asciidoc3_file)
        else:
            if not os.path.isdir(self.destination_dir):
                die('missing --destination-dir: %s' % self.destination_dir)
            self.destination_dir = os.path.abspath(self.destination_dir)
            if self.format not in ('chunked', 'epub', 'htmlhelp', 'xhtml'):
                warning('--destination-dir option is '
                        'only applicable to HTML based outputs')
        self.resource_dirs = []
        self.resource_files = []
        if self.resource_manifest:
            if not os.path.isfile(self.resource_manifest):
                die('missing --resource-manifest: %s' % self.resource_manifest)
            with open(self.resource_manifest) as resource_item:
                for elem in resource_item:
                    self.resources.append(elem.strip())
        for resource_candidate in self.resources:
            resource_candidate = os.path.expanduser(resource_candidate)
            resource_candidate = os.path.expandvars(resource_candidate)
            if resource_candidate.endswith('/') \
               or resource_candidate.endswith('\\'):
                if os.path.isdir(resource_candidate):
                    self.resource_dirs.append(resource_candidate)
                else:
                    die('missing resource directory: %s'
                        % resource_candidate)
            elif os.path.isdir(resource_candidate):
                self.resource_dirs.append(resource_candidate)
            elif (resource_candidate.startswith('.')) and \
                 ('=' in resource_candidate):
                ext, mimetype = resource_candidate.split('=')
                mimetypes.add_type(mimetype, ext)
            else:
                self.resource_files.append(resource_candidate)
        for conf_dir_file in (os.path.dirname(self.asciidoc3), CONF_DIR):
            for conf_alternative in ('images', 'stylesheets', 'docbook_v51'):
                conf_alternative = os.path.join(conf_dir_file,
                                                conf_alternative)
                if os.path.isdir(conf_alternative):
                    self.resource_dirs.append(conf_alternative)
        verbose('resource files: %s' % self.resource_files)
        verbose('resource directories: %s' % self.resource_dirs)
        if not self.doctype and self.format == 'manpage':
            self.doctype = 'manpage'
        if self.doctype:
            self.asciidoc3_opts += ' --doctype %s' % self.doctype
        for attr in self.attributes:
            self.asciidoc3_opts += ' --attribute "%s"' % attr
#        self.xsltproc_opts += ' --nonet'
        if self.verbose:
            self.asciidoc3_opts += ' --verbose'
            self.dblatex_opts += ' -V'
        if self.icons or self.icons_dir:
            params = [
                'callout.graphics 1',
                'navig.graphics 1',
                'admon.textlabel 0',
                'admon.graphics 1',
            ]
            if self.icons_dir:
                params += [
                    'admon.graphics.path "%s/"' % self.icons_dir,
                    'callout.graphics.path "%s/callouts/"' % self.icons_dir,
                    'navig.graphics.path "%s/"' % self.icons_dir,
                ]
        else:
            params = [
                'callout.graphics 0',
                'navig.graphics 0',
                'admon.textlabel 1',
                'admon.graphics 0',
            ]
        if self.stylesheet:
            params += ['html.stylesheet "%s"' % self.stylesheet]
        if self.format == 'htmlhelp':
            params += ['htmlhelp.chm "%s"' % self.basename('.chm'),
                       'htmlhelp.hhp "%s"' % self.basename('.hhp'),
                       'htmlhelp.hhk "%s"' % self.basename('.hhk'),
                       'htmlhelp.hhc "%s"' % self.basename('.hhc')]
        if self.doctype == 'book':
            params += ['toc.section.depth 1']
            # Books are chunked at chapter level.
            params += ['chunk.section.depth 0']
        for param_item in params:
            if param_item.split()[0]+' ' not in self.xsltproc_opts:
                self.xsltproc_opts += ' --stringparam ' + param_item
        if self.fop_opts:
            self.fop = True

    def dst_path(self, ext):
        '''
        Return name of file or directory in the destination directory with
        the same name as the asciidoc3 source file but with extension ext.
        '''
        return os.path.join(self.destination_dir, self.basename(ext))

    def basename(self, ext):
        '''
        Return the base name of the asciidoc3 source file but with extension
        ext.
        '''
        return os.path.basename(os.path.splitext(self.asciidoc3_file)[0]) + ext

    def asciidoc3_conf_file(self, path):
        '''
        Return full path name of file in asciidoc3 configuration files
        directory. Search first the directory containing the asciidoc3
        executable then the global configuration file directory.
        '''
        full_path_name = os.path.join(os.path.dirname(self.asciidoc3), path)
        if not os.path.isfile(full_path_name):
            full_path_name = os.path.join(CONF_DIR, path)
            if not os.path.isfile(full_path_name):
                die('missing configuration file: %s' % full_path_name)
        return os.path.normpath(full_path_name)

    def xsl_stylesheet(self, file_name=None):
        '''
        Return full path name of file in asciidoc3 docbook-xsl configuration
        directory.
        If an XSL file was specified with the --xsl-file option then it is
        returned.
        '''
        if self.xsl_file is not None:
            return self.xsl_file
        if not file_name:
            file_name = self.format + '.xsl'
        return self.asciidoc3_conf_file(os.path.join('docbook-xsl', file_name))

    def copy_resources(self, html_files, src_dir, dst_dir, resources=[]):
        '''
        Search html_files for images and CSS resource URIs (html_files can be a
        list of file names or a single file name).
        Copy them from the src_dir to the dst_dir.
        If not found in src_dir then recursively search all specified
        resource directories.
        Optional additional resources files can be passed
        in the resources list.
        '''
        resources = resources[:]
        # added in a2x3.py / AsciiDoc3:
        # see hints in docstring of function 'def find_resources()'
        resources += find_resources(html_files, 'link', 'href', lambda attrs:
                                    attrs.get('type') == 'text/css')
        resources += find_resources(html_files, 'img', 'src')
        resources += self.resource_files
        resources = sorted(set(resources))    # Drop duplicates.
        for resource_item in resources:
            if '=' in resource_item:
                src, dst = resource_item.split('=')
                if not dst:
                    dst = src
            else:
                src = dst = resource_item
            src = os.path.normpath(src)
            dst = os.path.normpath(dst)
            if os.path.isabs(dst):
                die('absolute resource file name: %s' % dst)
            if dst.startswith(os.pardir):
                die('resource file outside destination directory: %s' % dst)
            src = os.path.join(src_dir, src)
            dst = os.path.join(dst_dir, dst)
            if not os.path.isfile(src):
                for resource_dir_item in self.resource_dirs:
                    resource_dir_item = os.path.join(src_dir,
                                                     resource_dir_item)
                    found = find_files(resource_dir_item,
                                       os.path.basename(src))
                    if found:
                        src = found[0]
                        break
                else:
                    if not os.path.isfile(dst):
                        die('missing resource: %s' % src)
                    continue
            # Arrive here if resource file has been found.
            if os.path.normpath(src) != os.path.normpath(dst):
                dstdir = os.path.dirname(dst)
                shell_makedirs(dstdir)
                shell_copy(src, dst)

    def to_backend(self):
        '''
        Convert AsciiDoc3 source file to a backend output file using
        the global 'to_<backend name>' function (loaded from backend
        plugin a2x3-backend.py file).
        Executes the global function in an A2X3 class instance context.
        Previous versions have here
        'eval('to_%s(self)' % self.backend)'
        but we're trying to avoid dangerous 'eval()'.
        '''
        if self.backend == 'docbook':
            self.to_docbook(self)
        elif self.backend == 'xhtml':
            self.to_xhtml(self)
        elif self.backend == 'manpage':
            self.to_manpage(self)
        elif self.backend == 'pdf':
            self.to_pdf(self)
        elif self.backend == 'dvi':
            self.to_dvi(self)
        elif self.backend == 'ps':
            self.to_ps(self)
        elif self.backend == 'tex':
            self.to_tex(self)
        elif self.backend == 'htmlhelp':
            self.to_htmlhelp(self)
        elif self.backend == 'chunked':
            self.to_chunked(self)
        elif self.backend == 'epub':
            self.to_epub(self)
        elif self.backend == 'text':
            self.to_text(self)
        else:
            die('missing backend directive: {}'.format(self.backend))

    def to_docbook(self):
        '''
        Use asciidoc3 to convert asciidoc3_file to DocBook.
        args is a string containing additional asciidoc3 arguments.
        '''
        self.relaxng_location = relaxng_location
        self.use_dbv5 = False  # default = DocBook v45
        docbook_file = self.dst_path('.xml')
        if self.skip_asciidoc3:
            if not os.path.isfile(docbook_file):
                die('missing docbook file: %s' % docbook_file)
            return
        shell(
            '"%s" --backend docbook -a "a2x3-format=%s" %s '
            '--out-file "%s" "%s"'
            % (self.asciidoc3, self.format, self.asciidoc3_opts,
               docbook_file, self.asciidoc3_file))
        # #########
        # First, this is a workaround for the 'double title' bug, see
        # https://asciidoc3.org/blog/fixed-double-title-in-fop-generated-pdf.html
        # Second, this checks if 'docbook51' is used. If true, sets
        # appropriate xmllint options to validate.
        if os.name == 'nt' or os.name == 'posix' and not OPTIONS.dry_run:
            with open(docbook_file, "r") as fobj:
                xmltitlecheck = fobj.read()
            # pdt: alias for 'path_(to_)_double_title' to shorten linelength
            pdt = re.compile(
                r"""<((?:article|book|refentry)?)info>  # Tag *info group1
                    \s*                                 # opt. whitespace
                    (<.*>)                              # between tags group2
                    \s*                                 # opt. whitespace
                    \2                                  # group2 again!
                    \s*                                 # opt. whitespace
                    </\1info>                           # endtag *info group1 """,
                re.IGNORECASE | re.DOTALL | re.VERBOSE)
            xmltitlecheck, titlecounter = re.subn(pdt,
                                                  r'<\1info>\n\2\n</\1info>',
                                                  xmltitlecheck, 1)
            # Next line for debugging purposes only
            # if titlecounter: print("!!! Double Title !!!")
            if re.search(r'<.*?http://docbook\.org/ns/docbook.*?>',
                         xmltitlecheck):
                self.use_dbv5 = True
                if not re.search(r'<.*?version=.5.*?>', xmltitlecheck):
                    exit("""FATAL: DocBook v5 root element needs
                            an attribute 'version'! """)
            with open(docbook_file, "w") as fobj:
                fobj.write(xmltitlecheck)
        # End of workaround / DocBook5.x check.
        # #########
        if not self.no_xmllint and XMLLINT:
            # DocBook 5.1
            if self.use_dbv5:
                if not self.relaxng_location:
                    rng_path = '/tests/docbook_validation/rng-docbook51.rng'
                    if OPTIONS.resource_dirs:
                        try:
                            for resdirs_item in OPTIONS.resource_dirs:
                                self.relaxng_location = os.path.dirname(
                                    resdirs_item)
                                self.relaxng_location = \
                                    self.relaxng_location + rng_path
                                if os.path.isfile(self.relaxng_location):
                                    break
                        except OSError:
                            verbose(
                                """RelaxNG Schema for DocBook5.x not found.
                                   Please give full path in a2x3.py
                                   'relaxng_location'""")
                if not self.relaxng_location:
                    print("RelaxNG Schema for DocBook5.x not found.")
                    print("Please give path in a2x3.py 'relaxng_location' ")
                try:
                    shell('"%s" --noout --relaxng "%s" "%s"' %
                          (XMLLINT, self.relaxng_location, docbook_file))
                except OSError:
                    verbose("""No validation executed:
                               RelaxNG Schema for DocBook5.x not found. """)
            # DocBook 4.5
            else:
                shell('"%s" --nonet --noout --valid "%s"' %
                      (XMLLINT, docbook_file))

    def to_xhtml(self):
        """input.txt -> docbook.xml -> output.html """
        self.to_docbook()
        docbook_file = self.dst_path('.xml')
        xhtml_file = self.dst_path('.html')
        opts = '%s --output "%s"' % (self.xsltproc_opts, xhtml_file)
        exec_xsltproc(self.xsl_stylesheet(), docbook_file,
                      self.destination_dir, opts)
        src_dir = os.path.dirname(self.asciidoc3_file)
        self.copy_resources(xhtml_file,
                            src_dir, self.destination_dir)

    def to_manpage(self):
        """input.txt -> docbook.xml -> output.roff """
        self.to_docbook()
        docbook_file = self.dst_path('.xml')
        opts = self.xsltproc_opts
        exec_xsltproc(self.xsl_stylesheet(),
                      docbook_file, self.destination_dir, opts)

    def to_pdf(self):
        """if option --fop -> fop.pdf, else db_latex.pdf """
        if self.fop:
            self.exec_fop()
        else:
            self.exec_dblatex()

    def exec_fop(self):
        """input.txt -> docbook.xml -> pdf (using FOP) """
        self.to_docbook()
        docbook_file = self.dst_path('.xml')
        xsl = self.xsl_stylesheet('fo.xsl')
        fo_path = self.dst_path('.fo')
        pdf = self.dst_path('.pdf')
        opts = '%s --output "%s"' % (self.xsltproc_opts, fo_path)
        exec_xsltproc(xsl, docbook_file,
                      self.destination_dir, opts)
        shell('"%s" %s -fo "%s" -pdf "%s"' %
              (FOP, self.fop_opts, fo_path, pdf))
        if not self.keep_artifacts:
            shell_rm(fo_path)

    def exec_dblatex(self):
        """input.txt -> docbook.xml -> LaTeX.tex (dvi/ps/tex)"""
        self.to_docbook()
        docbook_file = self.dst_path('.xml')
        xsl = self.asciidoc3_conf_file(os.path.join(
            'dblatex', 'asciidoc3-dblatex.xsl'))
        sty = self.asciidoc3_conf_file(os.path.join(
            'dblatex', 'asciidoc3-dblatex.sty'))
        shell('"%s" -t %s -p "%s" -s "%s" %s "%s"' %
              (DBLATEX, self.format, xsl, sty,
               self.dblatex_opts, docbook_file))

    def to_dvi(self):
        """LaTeX output dvi"""
        self.exec_dblatex()

    def to_ps(self):
        """LaTeX output ps"""
        self.exec_dblatex()

    def to_tex(self):
        """LaTeX output tex"""
        self.exec_dblatex()

    def to_htmlhelp(self):
        """-> chunked"""
        self.to_chunked()

    def to_chunked(self):
        """input.txt -> docbook.xml -> xslt -> chunked/htmlhelp"""
        self.to_docbook()
        docbook_file = self.dst_path('.xml')
        opts = self.xsltproc_opts
        xsl_file = self.xsl_stylesheet()
        if self.format == 'chunked':
            dst_dir = self.dst_path('.chunked')
        elif self.format == 'htmlhelp':
            dst_dir = self.dst_path('.htmlhelp')
        if 'base.dir ' not in opts:
            opts += ' --stringparam base.dir "%s/"' % os.path.basename(dst_dir)
        # Create content.
        shell_rmtree(dst_dir)
        shell_makedirs(dst_dir)
        exec_xsltproc(xsl_file, docbook_file, self.destination_dir, opts)
        html_files = find_files(dst_dir, '*.html')
        src_dir = os.path.dirname(self.asciidoc3_file)
        self.copy_resources(html_files, src_dir, dst_dir)

    def update_epub_manifest(self, opf_file):
        """
        Scan the OEBPS (Open eBook Publication Structure) directory for any
        files that have not been registered in the OPF (Open Packaging Format)
        manifest then add them to the manifest.
        """
        opf_dir = os.path.dirname(opf_file)
        resource_files = []
        for (walk_item, __, files) in os.walk(os.path.dirname(opf_file)):
            for file_item in files:
                file_item = os.path.join(walk_item, file_item)
                if os.path.isfile(file_item):
                    assert file_item.startswith(opf_dir)
                    file_item = '.' + file_item[len(opf_dir):]
                    file_item = os.path.normpath(file_item)
                    if file_item not in ['content.opf']:
                        resource_files.append(file_item)
        opf = xml.dom.minidom.parseString(read_file(opf_file))
        manifest_files = []
        manifest = opf.getElementsByTagName('manifest')[0]
        for element in manifest.getElementsByTagName('item'):
            manifest_candidate = element.getAttribute('href')
            manifest_candidate = os.path.normpath(manifest_candidate)
            manifest_files.append(manifest_candidate)
        count = 0
        for resource_file_item in resource_files:
            if resource_file_item not in manifest_files:
                count += 1
                verbose('adding to manifest: %s' % resource_file_item)
                item = opf.createElement('item')
                item.setAttribute('href',
                                  resource_file_item.replace(os.path.sep,
                                                             '/'))
                item.setAttribute('id', 'a2x-%d' % count)
                mimetype = mimetypes.guess_type(resource_file_item)[0]
                if mimetype is None:
                    die('unknown mimetype: %s' % resource_file_item)
                item.setAttribute('media-type', mimetype)
                manifest.appendChild(item)
        if count > 0:
            write_file(opf_file, opf.toxml())

    def to_epub(self):
        """input.txt -> docbook.xml -> xsl
           -> epub/opf/html/zip (-> epubcheck)"""
        self.to_docbook()
        xsl_file = self.xsl_stylesheet()
        docbook_file = self.dst_path('.xml')
        epub_file = self.dst_path('.epub')
        build_dir = epub_file + '.d'
        shell_rmtree(build_dir)
        shell_makedirs(build_dir)
        # Create content.
        exec_xsltproc(xsl_file, docbook_file, build_dir, self.xsltproc_opts)
        # Copy resources referenced in the OPF and resources referenced by the
        # generated HTML (in theory DocBook XSL should ensure they are
        # identical but this is not always the case).
        src_dir = os.path.dirname(self.asciidoc3_file)
        dst_dir = os.path.join(build_dir, 'OEBPS')
        opf_file = os.path.join(dst_dir, 'content.opf')
        opf_resources = find_resources(opf_file, 'item', 'href')
        html_files = find_files(dst_dir, '*.html')
        self.copy_resources(html_files, src_dir, dst_dir, opf_resources)
        # Register any unregistered resources.
        self.update_epub_manifest(opf_file)
        # Build epub archive.
        cwd = os.getcwd()
        shell_cd(build_dir)
        try:
            if not self.dry_run:
                _zip = zipfile.ZipFile(epub_file, 'w')
                try:
                    # Create and add uncompressed mimetype file.
                    verbose('archiving: mimetype')
                    write_file('mimetype', 'application/epub+zip')
                    _zip.write('mimetype', compress_type=zipfile.ZIP_STORED)
                    # Compress all remaining files.
                    for (walk_path, __, files) in os.walk('.'):
                        for file_item in files:
                            file_item = os.path.normpath(
                                os.path.join(walk_path, file_item))
                            if file_item != 'mimetype':
                                verbose('archiving: %s' % file_item)
                                _zip.write(file_item,
                                           compress_type=zipfile.ZIP_DEFLATED)
                finally:
                    _zip.close()
            verbose('created archive: %s' % epub_file)
        finally:
            shell_cd(cwd)
        if not self.keep_artifacts:
            shell_rmtree(build_dir)
        if self.epubcheck and EPUBCHECK:
            if not find_executable(EPUBCHECK):
                warning('epubcheck skipped: unable to find executable: %s'
                        % EPUBCHECK)
            else:
                shell('"%s" "%s"' % (EPUBCHECK, epub_file))

    def to_text(self):
        """input.txt -> plain text via lynx or w3m"""
        text_file = self.dst_path('.text')
        html_file = self.dst_path('.text.html')
        if self.lynx:  # Option '-f text --lynx' given.
            shell(
                '"%s" %s --conf-file "%s" -b html4 '
                '-a "a2x3-format=%s" -o "%s" "%s"'
                % (self.asciidoc3, self.asciidoc3_opts,
                   self.asciidoc3_conf_file('text.conf'),
                   self.format, html_file, self.asciidoc3_file))
            shell('"%s" -dump "%s" > "%s"' %
                  (LYNX, html_file, text_file))
        else:
            # Use w3m(1), the default.
            self.to_docbook()
            docbook_file = self.dst_path('.xml')
            opts = '%s --output "%s"' % (self.xsltproc_opts, html_file)
            exec_xsltproc(self.xsl_stylesheet(), docbook_file,
                          self.destination_dir, opts)
            shell('"%s" -cols 70 -dump -T text/html -no-graph "%s" > "%s"' %
                  (W3M, html_file, text_file))
        if not self.keep_artifacts:
            shell_rm(html_file)


# ###################################################################
# Script main line.
# ###################################################################

def main():
    """ Reads cli options/arguments and starts execution."""

    global OPTIONS
    description = '''A toolchain manager for AsciiDoc3
                     (converts Asciidoc3 text files to other file formats)'''

    # arg_parser: we have already a var 'parser' in 'def find_resources()'
    arg_parser = ArgumentParser(usage='usage: %%prog [OPTIONS] SOURCE_FILE',
                                description=description)
    # options in alphabetical order
    arg_parser.add_argument('-a', '--attribute',
                            action='append', dest='attributes',
                            default=[], metavar='ATTRIBUTE',
                            help='set asciidoc3 attribute value')
    arg_parser.add_argument('--asciidoc3-opts',
                            action='append', dest='asciidoc3_opts',
                            default=[], metavar='ASCIIDOC3_OPTS',
                            help='asciidoc3 options')
    arg_parser.add_argument('-b', '--backend',
                            action='store', dest='backend', metavar='BACKEND',
                            help='name of backend plugin')
    arg_parser.add_argument('--backend-opts',
                            action='append', dest='backend_opts',
                            default=[], metavar='BACKEND_OPTS',
                            help='backend plugin options')
    arg_parser.add_argument('--conf-file', dest='conf_file',
                            default=None, metavar='primary_conf_file',
                            help='configuration file')
    # --copy: DEPRECATED
    arg_parser.add_argument('--copy', action='store_true',
                            dest='copy', default=False,
                            help='DEPRECATED: does nothing')
    arg_parser.add_argument('-D', '--destination-dir',
                            action='store', dest='destination_dir',
                            default=None, metavar='PATH',
                            help="""output directory
                                 (defaults to SOURCE_FILE directory)""")
    arg_parser.add_argument('-d', '--doctype',
                            action='store', dest='doctype', metavar='DOCTYPE',
                            choices=('article', 'manpage', 'book'),
                            help='article, manpage, book')
    arg_parser.add_argument('--dblatex-opts',
                            action='append', dest='dblatex_opts',
                            default=[], metavar='DBLATEX_OPTS',
                            help='dblatex options')
    arg_parser.add_argument('--epubcheck', action='store_true',
                            dest='epubcheck', default=False,
                            help='check EPUB output with epubcheck')
    arg_parser.add_argument('-f', '--format',
                            action='store', dest='format',
                            metavar='FORMAT', default='pdf',
                            choices=('chunked', 'epub', 'htmlhelp',
                                     'manpage', 'pdf', 'text', 'xhtml',
                                     'dvi', 'ps', 'tex', 'docbook'),
                            help='chunked, epub, htmlhelp, manpage, pdf, '
                            'text, xhtml, dvi, ps, tex, docbook')
    arg_parser.add_argument('--fop',
                            action='store_true', dest='fop', default=False,
                            help='use FOP to generate PDF files')
    arg_parser.add_argument('--fop-opts',
                            action='append', dest='fop_opts',
                            default=[], metavar='FOP_OPTS',
                            help='options for FOP pdf generation')
    arg_parser.add_argument('--icons',
                            action='store_true', dest='icons', default=False,
                            help="""use admonition, callout
                                    and navigation icons""")
    arg_parser.add_argument('--icons-dir',
                            action='store', dest='icons_dir',
                            default=None, metavar='PATH',
                            help='admonition and navigation icon directory')
    arg_parser.add_argument('-k', '--keep-artifacts', action='store_true',
                            dest='keep_artifacts', default=False,
                            help='do not delete temporary build files')
    arg_parser.add_argument('-L', '--no-xmllint', action='store_true',
                            dest='no_xmllint', default=False,
                            help="""do not check asciidoc3
                                    output with xmllint""")
    arg_parser.add_argument('--lynx',
                            action='store_true', dest='lynx', default=False,
                            help='use lynx to generate text files')
    arg_parser.add_argument('-m', '--resource-manifest',
                            action='store', dest='resource_manifest',
                            default=None, metavar='FILE',
                            help='read resources from FILE')
    arg_parser.add_argument('-n', '--dry-run', action='store_true',
                            dest='dry_run', default=False,
                            help="""just print the commands that would
                                    have been executed""")
    arg_parser.add_argument('-r', '--resource', action='append',
                            dest='resources', default=[], metavar='PATH',
                            help="""resource file or dir
                                    containing resource files""")
    # --resource-dir: DEPRECATED
    arg_parser.add_argument('--resource-dir',
                            action='append', dest='resources', default=[],
                            metavar='PATH', help='DEPRECATED: use --resource')
    # -s: DEPRECATED
    arg_parser.add_argument('-s', '--skip-asciidoc3',
                            action='store_true', dest='skip_asciidoc3',
                            default=False, help='DEPRECATED: redundant')
    # --safe: DEPRECATED
    arg_parser.add_argument('--safe',
                            action='store_true', dest='safe', default=False,
                            help='DEPRECATED: does nothing')
    arg_parser.add_argument('--stylesheet',
                            action='store', dest='stylesheet',
                            default=None, metavar='STYLESHEET',
                            help='HTML CSS stylesheet file name')
    arg_parser.add_argument('--version', action='version',
                            version='{PROG} {VERSION}'.format(PROG=PROG,
                                                              VERSION=VERSION))
    arg_parser.add_argument('--xsl-file',
                            action='store', dest='xsl_file',
                            metavar='XSL_FILE', help='custom XSL stylesheet')
    arg_parser.add_argument('--xsltproc-opts',
                            action='append', dest='xsltproc_opts',
                            default=[], metavar='XSLTPROC_OPTS',
                            help='xsltproc options for XSL stylesheets')
    arg_parser.add_argument('-v', '--verbose',
                            action='count', dest='verbose', default=0,
                            help='increase verbosity')
    # positional argument: a2x3/asciidoc3 input file
    arg_parser.add_argument("input_file", action="store",
                            help="AsciiDoc3 input file")
    source_options = get_source_options(sys.argv[-1])
    argv = source_options + sys.argv[1:]
    opts = arg_parser.parse_args(argv)
    opts.asciidoc3_opts = ' '.join(opts.asciidoc3_opts)
    opts.dblatex_opts = ' '.join(opts.dblatex_opts)
    opts.fop_opts = ' '.join(opts.fop_opts)
    opts.xsltproc_opts = ' '.join(opts.xsltproc_opts)
    opts.backend_opts = ' '.join(opts.backend_opts)
    # Convert argparse.Values to dict.
    # avoid eval() found in v2: 'opts = eval(str(opts))'
    opts = vars(opts)
    # Format 'epub' doesn't like option '--dry-run'.
    if opts['format'] == 'epub' and opts['dry_run']:
        print("Do not use option '--dry-run' together with '-f epub'!")
        sys.exit(1)
    a2x3 = A2X3(opts)
    OPTIONS = a2x3       # verbose and dry_run used by utility functions.
    verbose('args: %r' % argv)
    a2x3.asciidoc3_file = opts['input_file']
    try:
        a2x3.load_conf()
        a2x3.execute()
    except KeyboardInterrupt:
        sys.exit(1)


if __name__ == '__main__':
    main()



#!/usr/bin/env python3
"""
asciidoc3api - AsciiDoc3 API wrapper class.

The AsciiDoc3API class provides an API for executing asciidoc. Minimal example
compiles `mydoc.txt` to `mydoc.html`:

  import asciidoc3api
  asciidoc3 = asciidoc3api.AsciiDoc3API()
  asciidoc3.execute('mydoc.txt')

- Full documentation in asciidoc3api.txt.
- See the doctests below for more examples.

Doctests: ...

1. Check execution:

   >>> import io
   >>> infile = io.StringIO('Hello *{author}*')
   >>> outfile = io.StringIO()
   >>> asciidoc3 = AsciiDoc3API()
   >>> asciidoc3.options('--no-header-footer')
   >>> asciidoc3.attributes['author'] = 'Joe Bloggs'
   >>> asciidoc3.execute(infile, outfile, backend='html4')
   >>> print(outfile.getvalue())
   <p>Hello <strong>Joe Bloggs</strong></p>

   >>> asciidoc3.attributes['author'] = 'Bill Smith'
   >>> infile = io.StringIO('Hello _{author}_')
   >>> outfile = io.StringIO()
   >>> asciidoc3.execute(infile, outfile, backend='docbook')
   >>> print(outfile.getvalue())
   <simpara>Hello <emphasis>Bill Smith</emphasis></simpara>

2. Check error handling:

   >>> import io
   >>> asciidoc3 = AsciiDoc3API()
   >>> infile = io.StringIO('---------')
   >>> outfile = io.StringIO()
   >>> asciidoc3.execute(infile, outfile)
   Traceback (most recent call last):
     File "<stdin>", line 1, in <module>
     File "asciidoc3api.py", line 189, in execute
       raise AsciiDoc3Error(self.messages[-1])
   AsciiDoc3Error: ERROR: <stdin>: line 1: [blockdef-listing] missing closing delimiter

Copyright (C) 2009 Stuart Rackham.
Copyright (C) 2018-2020 Berthold Gehrke <berthold.gehrke@gmail.com>.
Free use of this software is granted under the terms of the
GNU General Public License v2 or higher (GPLv2+).
"""

import importlib
import os
from pathlib import Path
import sys

# next two lines are momentarily unused placeholders
# since we have no doctests yet
API_VERSION = '0.1.2'
MIN_ASCIIDOC3_VERSION = '3.1.0'


def find_in_path(fname, path=None):
    """
    Find file fname in paths. Return None if not found.
    """
    if path is None:
        path = os.environ.get('PATH', '')
    for _dir in path.split(os.pathsep):
        fpath = os.path.join(_dir, fname)
        if os.path.isfile(fpath):
            return fpath
    else:
        return None


class AsciiDoc3Error(Exception):
    """
    Helps to distinguish API-Errors from remaining Python-Errors.
    """
    pass


class Options(object):
    """
    Stores asciidoc3 command options.
    """

    def __init__(self, values=[]):
        self.values = values[:]

    def __call__(self, name, value=None):
        """Shortcut for append method."""
        self.append(name, value)

    def append(self, name, value=None):
        """append(optname, 4.56) appends tuple
           (optname, '4.56') to list 'values'"""
        if type(value) in (int, float):
            value = str(value)
        self.values.append((name, value))

# 'class Version' not yet implemented in AsciiDoc3
# used in Ascidoc v2 for doctests
# class Version(object):
#     """
#     Parse and compare AsciiDoc version numbers. Instance attributes:
#     ...
#     """
#     def __init__(self, version):
#         self.string = version
#         ...
#         return result


class AsciiDoc3API(object):
    """
    AsciiDoc3 API class.
    """
    def __init__(self, asciidoc3_py=None):
        """
        Locate and import asciidoc3.py.
        Initialize instance attributes.
        """
        self.options = Options()
        self.attributes = {}
        self.messages = []
        # Search for the asciidoc3 command file.
        # Try ASCIIDOC3_PY environment variable first.
        cmd = os.environ.get('ASCIIDOC3_PY')
        if cmd:
            if not os.path.isfile(cmd):
                raise AsciiDoc3Error('missing ASCIIDOC3_PY file: %s' % cmd)
        elif asciidoc3_py:
            # Next try path specified by caller.
            cmd = asciidoc3_py
            if not os.path.isfile(cmd):
                raise AsciiDoc3Error('missing file: %s' % cmd)
        else:
            # Try shell search paths.
            for fname in ['asciidoc3.py', 'asciidoc3.pyc', 'asciidoc3']:
                cmd = find_in_path(fname)
                if cmd:
                    break
            else:
                # Finally try current working directory.
                try:
                    # path_to_ad3: when copied via zip/tar.gz from gitlab
                    full_path = os.path.abspath(os.path.dirname(__file__))
                    path_to_ad3 = str(Path(full_path).parents[0])
                    for cmd in ['asciidoc3.py', 'asciidoc3.pyc',
                                'asciidoc3', path_to_ad3 + '/asciidoc3.py']:
                        if os.path.isfile(cmd):
                            break
                except:
                    raise AsciiDoc3Error('failed to locate asciidoc3')
        self.cmd = os.path.realpath(cmd)
        self.__import_asciidoc3()

    def __import_asciidoc3(self, reload=False):
        '''
        Import asciidoc3 module (script or compiled .pyc). See
        http://groups.google.com/group/asciidoc/browse_frm/thread/66e7b59d12cd2f91
        for an explanation of why a seemingly straight-forward job turned out
        quite complicated.
        '''
        if os.path.splitext(self.cmd)[1] in ['.py', '.pyc']:
            sys.path.insert(0, os.path.dirname(self.cmd))
            try:
                try:
                    if reload:
                        import builtins  # Because reload() is shadowed.
                        importlib.reload(self.asciidoc3)
                    else:
                        import asciidoc3
                        self.asciidoc3 = asciidoc3
                except ImportError:
                    raise AsciiDoc3Error('failed to import ' + self.cmd)
            finally:
                del sys.path[0]
        else:
            # 'else'-branch perhaps not neccessary any more in Python3
            # The import statement can only handle .py or .pyc files, have to
            # use imp.load_source() for scripts with other names.
            try:
                importlib.import_module('asciidoc3')
                self.asciidoc3 = asciidoc3
            except UnboundLocalError:
                self.asciidoc3 = importlib.import_module('asciidoc3')
            except ImportError:
                raise AsciiDoc3Error('failed to import ' + self.cmd)
#        (doctest) Not yet implemented in AsciiDoc3
#        if Version(self.asciidoc3.VERSION) < Version(MIN_ASCIIDOC3_VERSION):
#            raise AsciiDoc3Error(
#                'asciidoc3api %s requires asciidoc3 %s or better'
#                % (API_VERSION, MIN_ASCIIDOC3_VERSION))

    def execute(self, infile, outfile=None, backend=None):
        """
        Compile infile to outfile using backend format.
        infile and outfile can be file path strings or file like objects.
        """
        self.messages = []
        opts = Options(self.options.values)
        if outfile is not None:
            opts('--out-file', outfile)
        if backend is not None:
            opts('--backend', backend)
        for k, v in list(self.attributes.items()):
            if v == '' or k[-1] in '!@':
                s = k
            elif v is None:  # A None value undefines the attribute.
                s = k + '!'
            else:
                s = '%s=%s' % (k, v)
            opts('--attribute', s)
        args = [infile]
        # The AsciiDoc3 command was designed to process source text then
        # exit, there are globals and statics in asciidoc3.py that have
        # to be reinitialized before each run -- hence the reload.
        self.__import_asciidoc3(reload=True)
        try:
            try:
                self.asciidoc3.execute(self.cmd, opts.values, args)
            finally:
                self.messages = self.asciidoc3.messages[:]
        except SystemExit as e:
            if e.code:
                raise AsciiDoc3Error(self.messages[-1])


if __name__ == "__main__":
    # Run module doctests.
    import doctest
    options = doctest.NORMALIZE_WHITESPACE + doctest.ELLIPSIS
    doctest.testmod(optionflags=options)



#!/usr/bin/env python3

'''
./tests/docbook_validation/asciidoc3_docbook51_relaxng_validation.py
validation docbook51 relaxng

Tests all DocBook51-files produced with testasciidoc3.py found
in directory './tests/data' if valid against
relaxng_docbook51.
We are using the following five tools to validate:
- xmllint (Package libxml2-utils)
- xmlstarlet (Package xmlstarlet)
- lxml.de (Package python3-lxml)
- jing
- Sun's MSV
Check if present on command line with 'xmllint' and 'xmlstarlet' and
'jing' and with 'import lxml'.
You can disable the program that is not available, but it is recommended
to make use of more than one validator to make the test more reliable.

USAGE:
'python3 asciidoc3_docbook51_relaxng_validation.py'
to test all files ending with 'docbook51.xml' in directory './tests/data' or
'python3 asciidoc3_docbook51_relaxng_validation.py path/to/db51_xml_file.xml'
to validate just one file.

Copyright: (c) 2020 Berthold Gehrke <berthold.gehrke@gmail.com>
License:   GNU GPL v2 or higher
'''

from io import StringIO
import os
from pathlib import Path
import re
import subprocess
import sys
try:
    from lxml import etree
    LXMLPARSER = 'lxmlparser'
except ModuleNotFoundError:
    LXMLPARSER = ''

# External executables.
XMLLINT = 'xmllint'         # Set to '' to disable.
# XMLLINT = ''

XMLSTARLET = 'xmlstarlet'   # Set to '' to disable.
# XMLSTARLET = ''

JING = 'jing'   # Set to '' to disable.
# JING = ''

SUNMSV = 'sunmsv'   # Set to '' to disable.
# SUNMSV = ''

# LXMLPARSER = ''  # Uncomment to disable lmxlparser.


def xmllint_db51(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    subprocess.run(["xmllint", "--noout", "--relaxng",
                    full_path + "/rng-docbook51.rng", path_to_data + testfile])


def xmlstarlet_db51(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    subprocess.run(["xmlstarlet", "val", "--relaxng",
                    full_path + "/rng-docbook51.rng", path_to_data + testfile])


def jing_db51(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    print("jing: ")
    subprocess.run(["jing", full_path + "/rng-docbook51.rng",
                    path_to_data + testfile])


def sunmsv_db51(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    print("sunmsv: ")
    subprocess.run(["java", "-jar", full_path +
                    "/sun_msv_jar_files/relames-0.3.0.jar",
                    full_path + "/rng-docbook51.rng", path_to_data + testfile])


def lxml_db51(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    relaxng_new = etree.parse(full_path + "/rng-docbook51.rng")
    relaxng = etree.RelaxNG(relaxng_new)
    with open(path_to_data + testfile, 'r') as tfile:
        testdata = tfile.read()
        testdata, counter = re.subn(
            r'<\?xml version="1.0"(.*?)encoding="UTF-8"(.*?)\?>',
            r'<?xml version="1.0"\g<1>\g<2>?>', testdata, 1)
        assert counter == 1
        testdata = StringIO(testdata)
        t = etree.parse(testdata)
        relaxng.assertValid(t)
        result_exp = relaxng.validate(t)
        print("lxml.de validation against RelaxNG:  ", result_exp)


def main():
    if filepath:
        if XMLLINT:
            xmllint_db51(base_filepath)
        if XMLSTARLET:
            xmlstarlet_db51(base_filepath)
        if LXMLPARSER:
            lxml_db51(base_filepath)
    else:
        with os.scandir(path_to_data) as testpath:
            for entry in testpath:
                if entry.name.endswith('docbook51.xml'):
                    if XMLLINT:
                        xmllint_db51(entry.name)
                    if XMLSTARLET:
                        xmlstarlet_db51(entry.name)
                    if JING:
                        jing_db51(entry.name)
                    if SUNMSV:
                        sunmsv_db51(entry.name)
                    if LXMLPARSER:
                        lxml_db51(entry.name)


if __name__ == '__main__':
    filepath = ''
    full_path = os.path.abspath(os.path.dirname(__file__))
    if len(sys.argv) == 2:
        filepath = sys.argv[1]
        if filepath.endswith('/'):
            sys.exit("FATAL: Please enter path without trailing '/'!")
        base_filepath = os.path.basename(filepath)
        dirname = os.path.dirname(filepath)
        if dirname:
            path_to_data = dirname + "/"
        else:
            path_to_data = ""
    elif len(sys.argv) > 2:
        sys.exit("FATAL: Only one argument (path) possible!")
    elif len(sys.argv) == 1:
        path_to_tests = str(Path(full_path).parents[0])
        path_to_data = path_to_tests + "/data/"
    else:
        sys.exit("FATAL: Found no argument at all!")
    main()




#!/usr/bin/env python3

"""
Provides a test series. Many features of AsciiDoc3 are examined.
Use it together with testasciidoc3.conf and asciidoc3api.py
living in the same directory.
(c) Stuart Rackham (c) 2020 Berthold Gehrke <berthold.gehrke@gmail.com>.
Free use is granted under the
GNU General Public License v2 or higher (GPLv2+)
"""

import difflib
import getopt
import io
import os
import re
import sys

import asciidoc3api

USAGE = '''Usage: testasciidoc3.py [OPTIONS] COMMAND
Run AsciiDoc3 conformance tests specified in configuration FILE.

Commands:
  list                          List tests
  run [NUMBER] [BACKEND]        Execute tests
  update [NUMBER] [BACKEND]     Regenerate and update test data

Options:
  -f, --conf-file=CONF_FILE
        Use configuration file CONF_FILE (default configuration file is
        testasciidoc3.conf in testasciidoc3.py directory)
  --force
        Update all test data overwriting existing data'''


__version__ = '3.2.0'
__copyright__ = 'Copyright (C) 2009 Stuart Rackham, (c) 2020 Berthold Gehrke'

# Default backends:
BACKENDS = ('html4', 'xhtml11', 'docbook45', 'docbook51', 'html5')
BACKEND_EXT = {'html4': '.html', 'xhtml11': '.html', 'docbook45': '.xml',
               'docbook51': '.xml', 'slidy': '.html', 'html5': '.html'}


def iif(condition, iftrue, iffalse=None):
    """
    Immediate if c.f. ternary ?: operator.
    False value defaults to '' if the true value is a string.
    False value defaults to 0 if the true value is a number.
    """
    if iffalse is None:
        if isinstance(iftrue, str):
            iffalse = ''
        if isinstance(iftrue, (int, float)):
            iffalse = 0
    if condition:
        return iftrue
    return iffalse


def message(msg=''):
    """
    Output is written (only) to stderr.
    """
    print(msg, file=sys.stderr)


def strip_end(lines):
    """
    Strip blank strings from the end of list of strings.
    """
    for i in range(len(lines)-1, -1, -1):
        if not lines[i]:
            del lines[i]
        else:
            break


def normalize_data(lines):
    """
    Strip comments and trailing blank strings from lines.
    """
    result = [s for s in lines if not s.startswith('#')]
    strip_end(result)
    return result


class AsciiDoc3Test(object):
    """
    Serves as 'default' class, if no other configuration is given.
    """

    def __init__(self):
        self.number = None      # Test number (1..).
        self.name = ''          # Optional test name.
        self.title = ''         # Optional test name.
        self.description = []   # List of lines followoing title.
        self.source = None      # AsciiDoc3 test source file name.
        self.options = []
        self.attributes = {}
        self.backends = BACKENDS
        self.datadir = None     # Where output files are stored.
        self.disabled = False

    def backend_filename(self, backend):
        """
        Return the path name of the backend  output file that is generated from
        the test name and output file type.
        """
        return '%s-%s%s' % (
            os.path.normpath(os.path.join(self.datadir, self.name)),
            backend,
            BACKEND_EXT[backend])

    def parse(self, lines, confdir, datadir):
        """
        Parse conf file test section from list of text lines.
        """
        self.__init__()
        self.confdir = confdir
        self.datadir = datadir
        lines = Lines(lines)
        while not lines.eol():
            some_lines = lines.read_until(r'^%')
            if some_lines:
                if not some_lines[0].startswith('%'):
                    if some_lines[0][0] == '!':
                        self.disabled = True
                        self.title = some_lines[0][1:]
                    else:
                        self.title = some_lines[0]
                    self.description = some_lines[1:]
                    continue
                reo = re.match(r'^%\s*(?P<directive>[\w_-]+)', some_lines[0])
                if not reo:
                    raise ValueError
                directive = reo.groupdict()['directive']
                data = normalize_data(some_lines[1:])
                if directive == 'source':
                    if data:
                        self.source = os.path.normpath(os.path.join(
                            self.confdir, os.path.normpath(data[0])))
                elif directive == 'options':
                    self.options = eval(' '.join(data))
                    for i, v in enumerate(self.options):
                        if isinstance(v, str):
                            self.options[i] = (v, None)
                elif directive == 'attributes':
                    self.attributes = eval(' '.join(data))
                elif directive == 'backends':
                    self.backends = eval(' '.join(data))
                elif directive == 'name':
                    self.name = data[0].strip()
                else:
                    raise ValueError
        if not self.title:
            self.title = self.source
        if not self.name:
            self.name = os.path.basename(os.path.splitext(self.source)[0])

    def is_missing(self, backend):
        """
        Returns True if there is no output test data file for backend.
        """
        return not os.path.isfile(self.backend_filename(backend))

    def is_missing_or_outdated(self, backend):
        """
        Returns True if the output test data file is missing or out of date.
        """
        return self.is_missing(backend) or (
            os.path.getmtime(self.source)
            > os.path.getmtime(self.backend_filename(backend)))

    def get_expected(self, backend):
        """
        Return expected test data output for backend.
        """
        f = open(self.backend_filename(backend))
        try:
            result = f.readlines()
            # Strip line terminators.
            result = [s.rstrip() for s in result]
        finally:
            f.close()
        return result

    def generate_expected(self, backend):
        """
        Generate and return test data output for backend.
        """
        asciidoc3 = asciidoc3api.AsciiDoc3API()
        asciidoc3.options.values = self.options
        asciidoc3.attributes = self.attributes
        infile = self.source
        outfile = io.StringIO()
        asciidoc3.execute(infile, outfile, backend)
        # this is the db51 workaround handling <abstract>
        if backend == 'docbook51':
            # add default ('--doctype', 'article') if missing
            options_values_list = [item for item in asciidoc3.options.values]
            options_values_list = [item[0] for item in options_values_list]
            if '--doctype' not in options_values_list:
                asciidoc3.options.values.append(('--doctype', 'article'))
            if ('--doctype', 'article') in asciidoc3.options.values:
                input_so_far = outfile.getvalue()
                # abstract without preamble
                input_so_far, counter = re.subn(
                    r'</info>\s*(<abstract.*?</abstract>)',
                    r'\g<1>\n</info>', input_so_far, 1, flags=re.DOTALL)
                # print("counter", counter)  # debugging
                if not counter:
                    # abstract with preamble
                    input_so_far = re.sub(
                       r'</info>\s*(<simpara>.*?)\s?(<abstract.*?</abstract>)',
                       r'\g<2>\n</info>\n\g<1>',
                       input_so_far, 1, flags=re.DOTALL)
                return input_so_far.splitlines()
            if ('--doctype', 'book') in asciidoc3.options.values:
                input_so_far = outfile.getvalue()
                input_so_far, counter = re.subn(
                    r'</info>(.*?)(<abstract.*?</abstract>)\s?(.*?)<chapter',
                    r'\g<2>\n</info>\g<1>\g<3><chapter',
                    input_so_far, 1, flags=re.DOTALL)
                return input_so_far.splitlines()
        # end of db51 workaround
        return outfile.getvalue().splitlines()

    def update_expected(self, backend):
        """
        Generate and write backend data.
        """
        lines = self.generate_expected(backend)
        if not os.path.isdir(self.datadir):
            print(('CREATING: %s' % self.datadir))
            os.mkdir(self.datadir)
        f = open(self.backend_filename(backend), 'w+')
        try:
            print(('WRITING: %s' % f.name))
            f.writelines([s + os.linesep for s in lines])
        finally:
            f.close()

    def update(self, backend=None, force=False):
        """
        Regenerate and update expected test data outputs.
        """
        if backend is None:
            backends = self.backends
        else:
            backends = [backend]
        for backend in backends:
            if force or self.is_missing_or_outdated(backend):
                self.update_expected(backend)

    def run(self, backend=None):
        """
        Execute test.
        Return True if test passes.
        """
        if backend is None:
            backends = self.backends
        else:
            backends = [backend]
        result = True   # Assume success.
        self.passed = self.failed = self.skipped = 0
        print(('%d: %s' % (self.number, self.title)))
        if self.source and os.path.isfile(self.source):
            print(('SOURCE: asciidoc3: %s' % self.source))
            for backend in backends:
                fromfile = self.backend_filename(backend)
                if not self.is_missing(backend):
                    expected = self.get_expected(backend)
                    strip_end(expected)
                    got = self.generate_expected(backend)
                    strip_end(got)
                    lines = []
                    for line in difflib.unified_diff(got, expected, n=0):
                        lines.append(line)
                    if lines:
                        result = False
                        self.failed += 1
                        lines = lines[3:]
                        print(('FAILED: %s: %s' % (backend, fromfile)))
                        message('+++ %s' % fromfile)
                        message('--- got')
                        for line in lines:
                            message(line)
                        message()
                    else:
                        self.passed += 1
                        print(('PASSED: %s: %s' % (backend, fromfile)))
                else:
                    self.skipped += 1
                    print(('SKIPPED: %s: %s' % (backend, fromfile)))
        else:
            self.skipped += len(backends)
            if self.source:
                msg = 'MISSING: %s' % self.source
            else:
                msg = 'NO ASCIIDOC3 SOURCE FILE SPECIFIED'
            print(msg)
        print('')
        return result


class AsciiDoc3Tests(object):
    """
    Containing functions which 'do the work' of testing.
    """

    def __init__(self, conffile):
        """
        Parse configuration file.
        """
        self.conffile = os.path.normpath(conffile)
        # All file names are relative to configuration file directory.
        self.confdir = os.path.dirname(self.conffile)
        self.datadir = self.confdir  # Default expected files directory.
        self.tests = []  # List of parsed AsciiDoc3Test objects.
        self.globals = {}
        f = open(self.conffile)
        try:
            lines = Lines(f.readlines())
        finally:
            f.close()
        first = True
        while not lines.eol():
            s = lines.read_until(r'^%+$')
            s = [l for l in s if l]  # Drop blank lines.
            # Must be at least one non-blank line in addition to delimiter.
            if len(s) > 1:
                # Optional globals precede all tests.
                if first and re.match(r'^%\s*globals$', s[0]):
                    self.globals = eval(' '.join(normalize_data(s[1:])))
                    if 'datadir' in self.globals:
                        self.datadir = os.path.join(
                            self.confdir,
                            os.path.normpath(self.globals['datadir']))
                else:
                    test = AsciiDoc3Test()
                    test.parse(s[1:], self.confdir, self.datadir)
                    self.tests.append(test)
                    test.number = len(self.tests)
                first = False

    def run(self, number=None, backend=None):
        """
        Run all tests.
        If number is specified run test number (1..).
        """
        self.passed = self.failed = self.skipped = 0
        for test in self.tests:
            if (not test.disabled or number) and \
               (not number or number == test.number) and \
               (not backend or backend in test.backends):
                test.run(backend)
                self.passed += test.passed
                self.failed += test.failed
                self.skipped += test.skipped
        if self.passed > 0:
            print(('TOTAL PASSED:  %s' % self.passed))
        if self.failed > 0:
            print(('TOTAL FAILED:  %s' % self.failed))
        if self.skipped > 0:
            print(('TOTAL SKIPPED: %s' % self.skipped))

    def update(self, number=None, backend=None, force=False):
        """
        Regenerate expected test data and update configuration file.
        """
        for test in self.tests:
            if (not test.disabled or number) and \
               (not number or number == test.number):
                test.update(backend, force=force)

    def list(self):
        """
        Lists tests to stdout.
        """
        for test in self.tests:
            print('%d: %s%s' % (test.number, iif(test.disabled, '!'),
                                test.title))


class Lines(list):
    """
    A list of strings.
    Adds eol() and read_until() to list type.
    """

    def __init__(self, lines):
        super(Lines, self).__init__()
        self.extend([s.rstrip() for s in lines])
        self.pos = 0

    def eol(self):
        """
        Return 'True' if 'self.pos' is equal or greater than len(list).
        """
        return self.pos >= len(self)

    def read_until(self, regexp):
        """
        Return a list of lines from current position up until the next line
        matching regexp.
        Advance position to matching line.
        """
        result = []
        if not self.eol():
            result.append(self[self.pos])
            self.pos += 1
        while not self.eol():
            if re.match(regexp, self[self.pos]):
                break
            result.append(self[self.pos])
            self.pos += 1
        return result


def usage(msg=None):
    """
    Write 'msg' (if any) plus USAGE.
    """
    if msg:
        message(msg + '\n')
    message(USAGE)


if __name__ == '__main__':
    # Process command line options.
    # import getopt
    try:
        opts, args = getopt.getopt(sys.argv[1:], 'f:', ['force'])
    except getopt.GetoptError:
        usage('illegal command options')
        sys.exit(1)
    if not args:
        usage()
        sys.exit(1)
    conffile = os.path.join(os.path.dirname(sys.argv[0]), 'testasciidoc3.conf')
    force = False
    for o, v in opts:
        if o == '--force':
            force = True
        if o in ('-f', '--conf-file'):
            conffile = v
    if not os.path.isfile(conffile):
        message('missing CONF_FILE: %s' % conffile)
        sys.exit(1)
    tests = AsciiDoc3Tests(conffile)
    cmd = args[0]
    number = None
    backend = None
    for arg in args[1:3]:
        try:
            number = int(arg)
        except ValueError:
            backend = arg
    if backend and backend not in BACKENDS:
        message('illegal BACKEND: %s' % backend)
        sys.exit(1)
    if number is not None and \
       number not in list(range(1, len(tests.tests)+1)):
        message('illegal test NUMBER: %d' % number)
        sys.exit(1)
    if cmd == 'run':
        tests.run(number, backend)
        if tests.failed:
            exit(1)
    elif cmd == 'update':
        # dir 'images': cleanup
        if force:
            artifacts_to_remove = (
                'latex1.md5',
                'latex1.png',
                'latex-filter__1.md5',
                'latex-filter__1.png',
                'latex-filter__2.md5',
                'latex-filter__2.png',
                'latex-filter__3.md5',
                'latex-filter__3.png',
                'music1.md5',
                'music1.png',
                'music2.md5',
                'music2.png',
                'open-block-test__1.md5',
                'open-block-test_fop__1.md5',
                'open-block-test__1.png',
                'open-block-test__3.md5',
                'open-block-test_fop__3.md5',
                'open-block-test__3.png',
                'open-block-test_fop__1.png',
                'open-block-test_fop__2.png',
                'open-block-test_fop__3.png',
                'slidy-example__1.md5',
                'slidy-example__1.png',
                'graphviz1.png',
                'graphviz2.png',
                'open-block-test__1.png',
                'open-block-test__2.png',
                'open-block-test__3.png',
                'slidy-example__1.png',)
            for item in artifacts_to_remove:
                if os.path.exists("../images/" + item):
                    os.remove("../images/" + item)
        # end of cleanup
        tests.update(number, backend, force=force)
    elif cmd == 'list':
        tests.list()
    else:
        usage('illegal COMMAND: %s' % cmd)



#!/usr/bin/env python3

"""
Runs a bunch of a2x3.py testcases over alle txt-files in directory
./tests/data/
to produce PDFs via, first, dblatex and, secons, via Apache FOP. 

Usage: python3 a2x3_multitest.py

You'll see some errors because of not all documents are
valid DocBook. That's not a bug (nonvalid_docbook51.xml, newtables_docbook51
at al.), it's the expected behavior.

Copyright (C) 2020 by Berthold Gehrke <berthold.gehrke@gmail.com>
Free use of this software is granted under the terms of the
GNU General Public License Version 2 or higher (GNU GPLv2+).
"""

from concurrent import futures
import os
import subprocess
from shutil import copy


def a2x3_multi(testfile):
    '''
    TODO
    '''
    if testfile.endswith('_fop.txt'):
        subprocess.run(["python3", "../a2x3.py", "-f", "pdf", "--fop", path_to_data + testfile])
    else:
        subprocess.run(["python3", "../a2x3.py", "-f", "pdf", path_to_data + testfile])


def data_files():
    '''
    TODO
    '''
    testfiles = list()
    with os.scandir(path_to_data) as testpath:
        for entry in testpath:
            if entry.name.endswith('.txt'):
                testfiles.append(entry.name)
                filename_short = entry.name[:-4]
                copy('data/' + entry.name,
                     'data/' + filename_short + '_fop.txt')
                testfiles.append(filename_short + '_fop.txt')
    return tuple(testfiles)


def del_files():
    '''
    Remove files to have an "empty workspace"
    '''
    artifacts_to_remove = (
        'latex1.md5',
        'latex1.png',
        'latex-filter__1.md5',
        'latex-filter__1.png',
        'latex-filter__2.md5',
        'latex-filter__2.png',
        'latex-filter__3.md5',
        'latex-filter__3.png',
        'music1.md5',
        'music1.png',
        'music2.md5',
        'music2.png',
        'open-block-test__1.md5',
        'open-block-test_fop__1.md5',
        'open-block-test__1.png',
        'open-block-test__3.md5',
        'open-block-test_fop__3.md5',
        'open-block-test__3.png',
        'open-block-test_fop__1.png',
        'open-block-test_fop__2.png',
        'open-block-test_fop__3.png',
        'slidy-example__1.md5',
        'slidy-example__1.png',
        'graphviz1.png',
        'graphviz2.png',
        'open-block-test__1.png',
        'open-block-test__2.png',
        'open-block-test__3.png',
        'slidy-example__1.png',)
    for item in artifacts_to_remove:
        if os.path.exists("../images/" + item):
            os.remove("../images/" + item)
    with os.scandir(path_to_data) as testpath:
        for entry in testpath:
            if entry.name.endswith('_fop.txt'):
                os.remove(path_to_data + '/' + entry.name)
            elif entry.name.endswith('.pdf'):
                os.remove(path_to_data + '/' + entry.name)
            elif entry.name.endswith('.xml') and \
                 not entry.name.endswith('test-docbook45.xml') and \
                 not entry.name.endswith('test-docbook51.xml'):
                os.remove(path_to_data + '/' + entry.name)
            else:
                pass

def main(item):
    '''
    TODO
    '''
    print(item)
    a2x3_multi(item)


if __name__ == "__main__":
    '''do not run as an imported module or using an IDE like IDLE'''
    full_path = os.path.abspath(os.path.dirname(__file__))
    path_to_data = full_path + "/data/"
    del_files()
    # to remove only the 'test artifacts',
    # comment out the next three lines
    TUPLE_TEST = data_files()
    with futures.ProcessPoolExecutor(max_workers=4) as e:
        for item in TUPLE_TEST:
            e.submit(main, item)



#!/usr/bin/env python3

'''

PLACEHOLDER, see updates 
https://gitlab.com/asciidoc3/asciidoc3/-/blob/master/tests/docbook_validation/asciidoc3_docbook51_w3cxml_validation.py

./tests/docbook_validation/asciidoc3_docbook51_w3cxml_validation.py
validation docbook51 dtd

Tests all DocBook51-files produced with testasciidoc3.py found
in directory './tests/data' if valid against
usr/share/xml/docbook/schema/xsd/5.0/docbook.xsd


Copyright: (c) 2020 Berthold Gehrke <berthold.gehrke@gmail.com>
License:   GNU GPL v2 or higher

USAGE:
'python3 asciidoc3_docbook51_w3cxml_validation.py' to validate all '.*docbook51.xml' files in ./tests/data
'python3 asciidoc3_docbook51_w3cxml_validation.py /path/to/docbook51_file.xml'

'''

from io import StringIO
import os
from pathlib import Path
import re
import subprocess
import sys
try:
    from lxml import etree
    LXMLPARSER = 'lxmlparser'
except ModuleNotFoundError:
    print("FATAL: Module lxml not found!")
    print("Try 'sudo apt-install python3-lxml' or so ...")
    sys.exit()
   
#/usr/share/xml/docbook/schema/xsd/5.0/docbook.xsd

sys.exit("placeholder")





###!/usr/bin/env python3

"""
This script sets some symlinks inside the AsciiDoc3 package
installed via pip / pip3 from 'https://pypi.org/project/asciidoc3/'
with 'pip3 install --user asciidoc3'.
Run it immediately subsequently after install.
See https://asciidoc3.org/pypi.html for more information.
Do _not_ run as root/Admin - nothing happens.

Copyright (C) 2018-2020 by Berthold Gehrke <berthold.gehrke@gmail.com>
Free use of this software is granted under the terms of the
GNU General Public License Version 2 or higher (GNU GPLv2+).
"""

import os
import platform
import re
import shutil
import subprocess
import sys

def main():
    """
    Pip is of course available somce we just ran 'pip install'.
    Uses 'pip show asciidoc3' to locate AsciiDoc3's config-dirs.
    Then we set symlinks; at first within AsciiDoc3,
    then from the home directory to the data-files.
    This should work with POSIX systems like GNU/Linux and with Windows, too.
    """
    USERHOMEDIR = os.path.expanduser("~")
    # GNU/Linux: '/home/<username>', root: '/root'
    # Windows: 'C:\\Users\\<username>', Admin: 'C:\\Windows\\system32'

    def adjust_testad3():
        """ Adjust ./tests/testasciidoc3.py to run tests. """
        USERHOMEDIR_escaped = re.escape(USERHOMEDIR)
        if os.path.exists(USERHOMEDIR + r"\asciidoc3\tests\testasciidoc3.py"):
            with open(USERHOMEDIR + r"\asciidoc3\tests\testasciidoc3.py", 'r',
                      encoding='utf-8') as testfile:
                file_content = testfile.read()
                # Adjust AsciiDoc3API()
                file_content = re.sub(
                    r'asciidoc3 = asciidoc3api\.AsciiDoc3API\(\)',
                    r'# Change path if needed - see asciidoc3api.txt!\n'
                    r'        asciidoc3 = asciidoc3api.AsciiDoc3API(r"'
                    + USERHOMEDIR_escaped + r'aaasciidoc3aaasciidoc3.py")',
                    file_content, 1)
                # Adjust encoding (Windows has cp1252), part i
                file_content = re.sub(
                    r"f = open\(self\.backend_filename\(backend\)\)",
                    r"# Added encoding='utf-8', because Windows assumes cp1252\n"
                    r"        f = open(self.backend_filename(backend), encoding = 'utf-8')",
                    file_content, 1)
                # Adjust encoding (Windows has cp1252), part ii
                file_content = re.sub(
                    r"f = open\(self\.backend_filename\(backend\), 'w\+'\)",
                    r"# Added encoding='utf-8', because Windows assumes cp1252\n"
                    r"        f = open(self.backend_filename(backend), 'w+', encoding = 'utf-8')",
                    file_content, 1)
                file_content = re.sub(r'aaasciidoc3aaasciidoc3.py',
                                      r'\\asciidoc3\\asciidoc3.py',
                                      file_content, 1)
            with open(USERHOMEDIR + r"\asciidoc3\tests\testasciidoc3.py",
                      'w', encoding='utf-8') as testfile:
                testfile.write(file_content)


    ad3_location = ''
    try:
        runpip = subprocess.Popen("pip show asciidoc3", shell=True,
                                  stdin=subprocess.PIPE,
                                  stdout=subprocess.PIPE,
                                  universal_newlines=True, bufsize=-1)
        runpip.wait()
        pip_show_status = runpip.returncode
        if pip_show_status:
            sys.exit("FATAL: 'pip show asciidoc3' ends with non-zero exit code")
        output = runpip.communicate()
    except OSError:
        sys.exit("FATAL: 'pip show asciidoc3' ends with an unidentified error ...")
    if output:
        output = output[0]
        output = re.split(r'Location: ', output, re.DOTALL)[1]
        output = re.split(r'\nRequires', output, re.DOTALL)[0]
        ad3_location = output
        #  string 'ad3_location' looks like this:
        #  GNU/Linux '/home/<username>/.local/lib/python3.5/site-packages'
        #  Windows 'c:\\users\\<username>\\appdata\\roaming\\python\\python37\\site-packages'
        #  Windows venv 'c:\\users\\<username>\\<venv-dir>\\lib\\site-packages'
    else:
        sys.exit("FATAL: 'pip show asciidoc3' gives no information ...")

    # ###########################
    # POSIX (GNU/Linux, BSD, ...)
    # ###########################
    if os.name == 'posix':
        if USERHOMEDIR.lower().startswith(r'/root'):
            sys.exit("FATAL: Do not run as root! Script does nothing at all.")

        # AsciiDoc3 distribution, set internal (relative) symlinks.
        # <...>/asciidoc3/filters/music/images --> ../../images
        if os.path.exists(ad3_location +r"/asciidoc3/filters/music/images"):
            os.unlink(ad3_location +r"/asciidoc3/filters/music/images")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/images", ad3_location + r"/asciidoc3/filters/music"),
                   ad3_location + r"/asciidoc3/filters/music/images")

        # <...>/asciidoc3/filters/graphviz/images --> ../../images
        if os.path.exists(ad3_location + r"/asciidoc3/filters/graphviz/images"):
            os.unlink(ad3_location + r"/asciidoc3/filters/graphviz/images")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/images", ad3_location + r"/asciidoc3/graphviz/music"),
                   ad3_location + r"/asciidoc3/filters/graphviz/images")

        # <...>/asciidoc3/doc/images --> ../images
        if os.path.exists(ad3_location + r"/asciidoc3/doc/images"):
            os.unlink(ad3_location + r"/asciidoc3/doc/images")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/images", ad3_location + r"/asciidoc3/doc"),
                   ad3_location + r"/asciidoc3/doc/images")

        # <...>/asciidoc3/tests/images --> ../images
        if os.path.exists(ad3_location + r"/asciidoc3/tests/images"):
            os.unlink(ad3_location + r"/asciidoc3/tests/images")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/images", ad3_location + r"/asciidoc3/tests"),
                   ad3_location + r"/asciidoc3/tests/images")

        # <...>/asciidoc3/tests/data/images --> ../../images
        if os.path.exists(ad3_location + r"/asciidoc3/tests/data/images"):
            os.unlink(ad3_location + r"/asciidoc3/tests/data/images")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/images", ad3_location + r"/asciidoc3/tests/data"),
                   ad3_location + r"/asciidoc3/tests/data/images")

        # <...>/asciidoc3/tests/asciidoc3api.py --> ../asciidoc3api.py
        if os.path.exists(ad3_location + r"/asciidoc3/tests/asciidoc3api.py"):
            os.unlink(ad3_location + r"/asciidoc3/tests/asciidoc3api.py")
        os.symlink(os.path.relpath(
            ad3_location + r"/asciidoc3/asciidoc3api.py", ad3_location + r"/asciidoc3/tests"),
                   ad3_location + r"/asciidoc3/tests/asciidoc3api.py")

        # Set symlinks users home to AsciiDoc3's 'working/config' directories
        # Skip if running as root.
        if USERHOMEDIR != r'/root':
            if os.path.exists(USERHOMEDIR + r"/.asciidoc3"):
                os.replace(USERHOMEDIR + r"/.asciidoc3", USERHOMEDIR + r"/.asciidoc3_backup")
            os.symlink(ad3_location + r"/asciidoc3", USERHOMEDIR + r"/.asciidoc3")

            if os.path.exists(USERHOMEDIR + r"/asciidoc3"):
                os.replace(USERHOMEDIR + r"/asciidoc3", USERHOMEDIR + r"/asciidoc3_backup")
            os.symlink(ad3_location + r"/asciidoc3", USERHOMEDIR + r"/asciidoc3")

        # Adjust path in ./tests/testasciidoc3.py to run tests
        if os.path.exists(ad3_location +r"/asciidoc3/tests/testasciidoc3.py"):
            with open(ad3_location +r"/asciidoc3/tests/testasciidoc3.py", 'r') as testfile:
                file_content = testfile.read()
                file_content = re.sub(r'asciidoc3 = asciidoc3api\.AsciiDoc3API\(\)',
                                      r'# Change path if needed - see ./doc/asciidoc3api.txt!\n'
                                      r'        asciidoc3 = asciidoc3api.AsciiDoc3API("'
                                      + ad3_location + r'/asciidoc3/asciidoc3.py")',
                                      file_content, 1)
            with open(ad3_location +r"/asciidoc3/tests/testasciidoc3.py", 'w') as testfile:
                testfile.write(file_content)

    # #############################################
    # Win8, Win10 (Win7 works, but has reached EOL) 
    # #############################################
    elif platform.system().lower().startswith(r"win"):
        if USERHOMEDIR.lower().startswith(r'c:\win'):
            sys.exit("FATAL: Do not run as ADMIN! Script does nothing at all.")

        # not using venv -> ad3_location looks like this:
        # 'c:\\users\\<username>\\appdata\\roaming\\python\\python37\\site-packages'
        # and we have a 'Scripts'-directory here:
        # 'c:\\users\\<username>\\appdata\\roaming\\python\\python37\\Scripts\'
        if ad3_location.count(r'appdata'):
            # we need modules in dir 'Scripts', that means sources without trailing '.py'
            shutil.copy(ad3_location[:-13]+r"Scripts\asciidoc3.exe", \
                        ad3_location[:-13]+r"Scripts\asciidoc3")
            shutil.copy(ad3_location[:-13]+r"Scripts\a2x3.exe", \
                        ad3_location[:-13]+r"Scripts\a2x3")

            # AsciiDoc3 distribution, set internal symlinks.
            # <...>/asciidoc3/filters/music/images --> ../../images
            if os.path.exists(ad3_location + r"\asciidoc3\filters\music\images"):
                os.unlink(ad3_location + r"\asciidoc3\filters\music\images")
            os.system(r"mklink /J "+ad3_location+r"\asciidoc3\filters\music\images "+ad3_location+r"\asciidoc3\images")

            # <...>/asciidoc3/filters/graphviz/images --> ../../images
            if os.path.exists(ad3_location + r"\asciidoc3\filters\graphviz\images"):
                os.unlink(ad3_location + r"\asciidoc3\filters\graphviz\images")
            os.system(r"mklink /J "+ad3_location+r"\asciidoc3\filters\graphviz\images "+ad3_location+r"\asciidoc3\images")

            # <...>/asciidoc3/doc/images --> ../images
            if os.path.exists(ad3_location + r"\asciidoc3\doc\images"):
                os.unlink(ad3_location + r"\asciidoc3\doc\images")
            os.system(r"mklink /J "+ad3_location+r"\asciidoc3\doc\images "+ad3_location+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/images --> ../images
            if os.path.exists(ad3_location + r"\asciidoc3\tests\images"):
                os.unlink(ad3_location + r"\asciidoc3\tests\images")
            os.system(r"mklink /J "+ad3_location+r"\asciidoc3\tests\images "+ad3_location+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/data/images --> ../../images
            if os.path.exists(ad3_location + r"\asciidoc3\tests\data\images"):
                os.unlink(ad3_location + r"\asciidoc3\tests\data\images")
            os.system(r"mklink /J "+ad3_location+r"\asciidoc3\tests\data\images "+ad3_location+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/asciidoc3api.py --> ../asciidoc3api.py
            if os.path.exists(ad3_location + r"\asciidoc3\tests\asciidoc3api.py"):
                os.unlink(ad3_location + r"\asciidoc3\tests\asciidoc3api.py")
            os.system(r"mklink /H "+ad3_location+r"\asciidoc3\tests\asciidoc3api.py "+ad3_location+r"\asciidoc3\asciidoc3api.py")

            # make symlink for convenient use
            if os.path.exists(USERHOMEDIR + r"\asciidoc3"):
                if os.path.exists(USERHOMEDIR + r"\asciidoc3_backup"):
                    shutil.rmtree(USERHOMEDIR + r"\asciidoc3_backup")
                os.replace(USERHOMEDIR + r"\asciidoc3", USERHOMEDIR + r"\asciidoc3_backup")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3 "+ad3_location+r"\asciidoc3")

            adjust_testad3()
            
            # make symlink .asciidoc3: mandatory since AsciiDoc3 looks here for config-files
            if os.path.exists(USERHOMEDIR + r"\.asciidoc3"):
                shutil.rmtree(USERHOMEDIR + r"\.asciidoc3")
            os.system(r"mklink /J "+USERHOMEDIR+r"\.asciidoc3 "+ad3_location+r"\asciidoc3")

            # 'c:\\users\\<username>\\appdata\\roaming\\python\\asciidoc3'
            # we don't need this folder any more
            if os.path.exists(ad3_location[:-22]+r"asciidoc3"):
                shutil.rmtree(ad3_location[:-22]+r"asciidoc3")

        # using venv -> ad3_location looks like this:
        # 'c:\\users\\username\\ad3\\lib\\site-packages'
        elif 'lib' in ad3_location.lower():
            # delete dir '..\\site-packages\\asciidoc3'
            for folder in os.listdir(ad3_location[:-17]):
                if folder.lower().startswith('python3'):
                    if os.path.exists(ad3_location[:-17]+folder+r'\site-packages\asciidoc3'):
                        shutil.rmtree(ad3_location[:-17]+folder+r'\site-packages\asciidoc3')
                        if len(list(os.listdir(ad3_location[:-17]+folder))) == 1:
                            shutil.rmtree(ad3_location[:-17]+folder+r'\site-packages')
                            shutil.rmtree(ad3_location[:-17]+folder)

            # we need modules in dir 'Scripts', that means sources without trailing '.py'
            shutil.copy(ad3_location[:-17]+r"Scripts\asciidoc3.exe", \
                        ad3_location[:-17]+r"Scripts\asciidoc3")
            shutil.copy(ad3_location[:-17]+r"Scripts\a2x3.exe", \
                        ad3_location[:-17]+r"Scripts\a2x3")

            # move folder asciidoc3 to apropriate location
            shutil.move(ad3_location[:-17] + r"\asciidoc3", ad3_location + r"\asciidoc3")

            # make symlink <venv>\asciidoc3
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3 "+ad3_location+r"\asciidoc3")

            # AsciiDoc3 distribution, set internal symlinks.
            # <...>/asciidoc3/filters/music/images --> ../../images
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\filters\music\images"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\filters\music\images")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3\filters\music\images "+USERHOMEDIR+r"\asciidoc3\images")

            # <...>/asciidoc3/filters/graphviz/images --> ../../images
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\filters\graphviz\images"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\filters\graphviz\images")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3\filters\graphviz\images "+USERHOMEDIR+r"\asciidoc3\images")

            # <...>/asciidoc3/doc/images --> ../images
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\doc\images"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\doc\images")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3\doc\images "+USERHOMEDIR+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/images --> ../images
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\tests\images"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\tests\images")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3\tests\images "+USERHOMEDIR+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/data/images --> ../../images
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\tests\data\images"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\tests\data\images")
            os.system(r"mklink /J "+USERHOMEDIR+r"\asciidoc3\tests\data\images "+USERHOMEDIR+r"\asciidoc3\images")

            # <...>/asciidoc3/tests/asciidoc3api.py --> ../asciidoc3api.py
            if os.path.exists(USERHOMEDIR + r"\asciidoc3\tests\asciidoc3api.py"):
                os.unlink(USERHOMEDIR + r"\asciidoc3\tests\asciidoc3api.py")
            os.system(r"mklink /H "+USERHOMEDIR+r"\asciidoc3\tests\asciidoc3api.py "+USERHOMEDIR+r"\asciidoc3\asciidoc3api.py")

            adjust_testad3()

            # make symlink .asciidoc3: mandatory since AsciiDoc3 looks here for config-files
            if os.path.exists(USERHOMEDIR + r"\.asciidoc3"):
                shutil.rmtree(USERHOMEDIR + r"\.asciidoc3")
            os.system(r"mklink /J " + USERHOMEDIR + r"\.asciidoc3 "
                      + ad3_location + r"\asciidoc3")

            # make copy <venv>\Scripts python - python3
            if not os.path.exists(ad3_location[:-17] + r"Scripts\python3.exe"):
                os.system(r"copy /Y "+ad3_location[:-17] + r"Scripts\python.exe "
                          + ad3_location[:-17] + r"Scripts\python3.exe")

    # ##########################
    # # neither POSIX or Windows
    # ##########################
    else:
        sys.exit("""\n
                 Your Operating System was not included in this script:
                 you are on your own to handle symlinks ...""")


if __name__ == '__main__':
    main()



#!/usr/bin/env python3

"""
This script 'setup' builds the AsciiDoc3 package to be installed via
pip / pip3 from 'https://pypi.org/project/asciidoc3/'.
Pip makes AsciiDoc3 available both on GNU/Linux (POSIX) and
Windows.
To complete the installation it is strongly recommended (or to say
it in a more accurate word: compelling) to run 'asciidoc3_postinstall'
from the command line immediately subsequently after
'pip3 install --user asciidoc3'. This arranges some reasonable
symlinks for convenient usage.
See https://asciidoc3.org/pypi.html for more information.

Copyright (C) 2018-2020 by Berthold Gehrke <berthold.gehrke@gmail.com>
Free use of this software is granted under the terms of the
GNU General Public License Version 2 or higher (GNU GPLv2+).
"""

from os import name
from sys import version
from setuptools import setup, find_packages

# find current version and location of executable
# e.g.: version = '3.7.2+ (default, Feb  2 2019, 14:31:48) \n[GCC 8.2.0]'
# PRE = '37'
# PRE='Python37/site-packages/'
PRE = version[:1] + version[2:3]
PRE = 'Python' + PRE + '/site-packages/'

# this is to assure right-installing on Windows *and* GNU/Linux
PREFIX_TUPLE = tuple()
if name == 'nt':
    PREFIX_TUPLE = ('', PRE)
elif name == 'posix':
    PREFIX_TUPLE = ('/',)
else:  # guess
    PREFIX_TUPLE = ('/',)

datafiles = list()
for dirprefix in PREFIX_TUPLE:
    datafiles.append((dirprefix+'asciidoc3',
                      ['asciidoc3.conf',
                       'COPYING',
                       'COPYRIGHT',
                       'docbook45.conf',
                       'docbook51.conf',
                       'help.conf',
                       'html4.conf',
                       'html5.conf',
                       'lang-de.conf',
                       'latex.conf',
                       'asciidoc3.py',
                       'a2x3.py',
                       'asciidoc3api.py',
                       'LICENSE',
                       'README.md',
                       'slidy.conf',
                       'setup.py',
                       'text.conf',
                       'xhtml11.conf',
                       'BUGS.txt',
                       'CHANGELOG',
                       'CHANGELOG.txt',
                       'COPYING',
                       'COPYRIGHT',
                       'INSTALL',
                       'UNINSTALL',
                       'lang-cs.conf',
                       'lang-el.conf',
                       'lang-es.conf',
                       'lang-fr.conf',
                       'lang-hu.conf',
                       'lang-it.conf',
                       'lang-ja.conf',
                       'lang-nl.conf',
                       'lang-pt-BR.conf',
                       'lang-ro.conf',
                       'lang-ru.conf',
                       'lang-se.conf',
                       'lang-zh-CN.conf',
                       'lang-uk.conf',
                       'lang-en.conf',
                       'asciidoc3_postinstall.py', ]))
    datafiles.append((dirprefix+'asciidoc3/stylesheets',
                      ['stylesheets/asciidoc3.css',
                       'stylesheets/docbook-xsl.css',
                       'stylesheets/pygments.css',
                       'stylesheets/slidy.css',
                       'stylesheets/toc2.css', ]))
    datafiles.append((dirprefix+'asciidoc3/dblatex',
                      ['dblatex/asciidoc3-dblatex.sty',
                       'dblatex/asciidoc3-dblatex.xsl',
                       'dblatex/dblatex-readme.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/docbook-xsl',
                      ['docbook-xsl/asciidoc3-docbook-xsl.txt',
                       'docbook-xsl/chunked.xsl',
                       'docbook-xsl/common.xsl',
                       'docbook-xsl/epub.xsl',
                       'docbook-xsl/fo.xsl',
                       'docbook-xsl/htmlhelp.xsl',
                       'docbook-xsl/manpage.xsl',
                       'docbook-xsl/text.xsl',
                       'docbook-xsl/xhtml.xsl', ]))
    datafiles.append((dirprefix+'asciidoc3/images',
                      ['images/1.png',
                       'images/2.png',
                       'images/3.png',
                       'images/empty.png',
                       'images/helloworld.jpg',
                       'images/highlighter.png',
                       'images/highlight.jpg',
                       'images/logo_asciidoc3.png',
                       'images/redsquare.jpg',
                       'images/smallnew.png',
                       'images/tiger.png', ]))
    datafiles.append((dirprefix+'asciidoc3/images/icons',
                      ['images/icons/caution.png',
                       'images/icons/example.png',
                       'images/icons/home.png',
                       'images/icons/important.png',
                       'images/icons/next.png',
                       'images/icons/note.png',
                       'images/icons/prev.png',
                       'images/icons/README',
                       'images/icons/tip.png',
                       'images/icons/up.png',
                       'images/icons/warning.png', ]))
    datafiles.append((dirprefix+'asciidoc3/images/icons/callouts',
                      ['images/icons/callouts/1.png',
                       'images/icons/callouts/2.png',
                       'images/icons/callouts/3.png',
                       'images/icons/callouts/4.png',
                       'images/icons/callouts/5.png',
                       'images/icons/callouts/6.png',
                       'images/icons/callouts/7.png',
                       'images/icons/callouts/8.png',
                       'images/icons/callouts/9.png',
                       'images/icons/callouts/10.png',
                       'images/icons/callouts/11.png',
                       'images/icons/callouts/12.png',
                       'images/icons/callouts/13.png',
                       'images/icons/callouts/14.png',
                       'images/icons/callouts/15.png', ]))
    datafiles.append((dirprefix+'asciidoc3/doc',
                      ['doc/asciidoc3api.txt',
                       'doc/asciidoc3.conf',
                       'doc/customers.csv',
                       'doc/article.txt',
                       'doc/article_docbook51.txt',
                       'doc/article-docinfo.xml',
                       'doc/article_docbook51-docinfo.xml',
                       'doc/asciimathml.txt',
                       'doc/book-multi.txt',
                       'doc/book-multi_docbook51.txt',
                       'doc/book.txt',
                       'doc/book_containing_an_abstract.txt',
                       'doc/cheatsheet.txt',
                       'doc/epub-notes.txt',
                       'doc/faq.txt',
                       'doc/latex-backend.txt',
                       'doc/latex-bugs.txt',
                       'doc/latex-filter.txt',
                       'doc/latexmathml.txt',
                       'doc/latexmath.txt',
                       'doc/music-filter.txt',
                       'doc/publishing-ebooks-with-asciidoc3.txt',
                       'doc/quickstart.txt',
                       'doc/readme.txt',
                       'doc/readme_docbook51.txt',
                       'doc/slidy-example.txt',
                       'doc/slidy.txt',
                       'doc/source-highlight-filter.txt',
                       'doc/test.txt',
                       'doc/a2x3.1.gz',
                       'doc/a2x3.1.txt',
                       'doc/asciidoc3.1.gz',
                       'doc/asciidoc3.1.txt',
                       'doc/userguide.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/javascripts',
                      ['javascripts/asciidoc3.js',
                       'javascripts/ASCIIMathML.js',
                       'javascripts/LaTeXMathML.js',
                       'javascripts/slidy.js',
                       'javascripts/toc.js', ]))
    datafiles.append((dirprefix+'asciidoc3/filters/code',
                      ['filters/code/code-filter.conf',
                       'filters/code/code-filter.py',
                       'filters/code/code-filter-readme.txt',
                       'filters/code/code-filter-test.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/filters/graphviz',
                      ['filters/graphviz/asciidoc3-graphviz-sample.txt',
                       'filters/graphviz/graphviz2png.py',
                       'filters/graphviz/graphviz-filter.conf', ]))
    datafiles.append((dirprefix+'asciidoc3/filters/latex',
                      ['filters/latex/latex2png.py',
                       'filters/latex/latex-filter.conf', ]))
    datafiles.append((dirprefix+'asciidoc3/filters/music',
                      ['filters/music/music-filter.conf',
                       'filters/music/example_music-filter.txt',
                       'filters/music/music-filter-test.txt',
                       'filters/music/music2png.py', ]))
    datafiles.append((dirprefix+'asciidoc3/filters/source',
                      ['filters/source/source-highlight-filter.conf',
                       'filters/source/source-highlight-filter-test.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/themes/flask',
                      ['themes/flask/flask.css', ]))
    datafiles.append((dirprefix+'asciidoc3/themes/volnitsky',
                      ['themes/volnitsky/volnitsky.css', ]))
    datafiles.append((dirprefix+'asciidoc3/vim',
                      ['vim/readme-vim.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/vim/syntax',
                      ['vim/syntax/asciidoc3.vim', ]))
    datafiles.append((dirprefix+'asciidoc3/man',
                      ['doc/a2x3.1.gz',
                       'doc/a2x3.1.txt',
                       'doc/asciidoc3.1.gz',
                       'doc/asciidoc3.1.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/tests',
                      ['tests/a2x3_multitest.py',
                       'tests/readme-tests.txt',
                       'tests/testa2x3.py',
                       'tests/testasciidoc3.conf',
                       'tests/testasciidoc3.py', ]))
    datafiles.append((dirprefix+'asciidoc3/tests/data',
                      ['tests/data/customers.csv',
                       'tests/data/testcases_docbook45.conf',
                       'tests/data/testcases_docbook51.conf',
                       'tests/data/lang-cs-test.txt',
                       'tests/data/lang-en-test.txt',
                       'tests/data/lang-fr-test.txt',
                       'tests/data/lang-it-test.txt',
                       'tests/data/lang-pt-BR-test.txt',
                       'tests/data/lang-ru-test.txt',
                       'tests/data/newtables.txt',
                       'tests/data/newtables_docbook51.txt',
                       'tests/data/utf8-examples.txt',
                       'tests/data/filters-test.txt',
                       'tests/data/lang-de-man-test.txt',
                       'tests/data/lang-es-man-test.txt',
                       'tests/data/lang-hu-man-test.txt',
                       'tests/data/lang-nl-man-test.txt',
                       'tests/data/lang-ro-man-test.txt',
                       'tests/data/lang-uk-man-test.txt',
                       'tests/data/nonvalid_docbook51.txt',
                       'tests/data/testcases_docbook45.txt',
                       'tests/data/testcases_docbook51.txt',
                       'tests/data/lang-de-test.txt',
                       'tests/data/lang-es-test.txt',
                       'tests/data/lang-hu-test.txt',
                       'tests/data/lang-ja-test.txt',
                       'tests/data/lang-nl-test.txt',
                       'tests/data/lang-ro-test.txt',
                       'tests/data/lang-se-test.txt',
                       'tests/data/lang-uk-test.txt',
                       'tests/data/open-block-test.txt',
                       'tests/data/lang-cs-man-test.txt',
                       'tests/data/lang-en-man-test.txt',
                       'tests/data/lang-fr-man-test.txt',
                       'tests/data/lang-it-man-test.txt',
                       'tests/data/lang-pt-BR-man-test.txt',
                       'tests/data/lang-ru-man-test.txt',
                       'tests/data/rcs-id-marker-test.txt',
                       'tests/data/utf8-bom-example.txt',
                       'tests/data/utf8-bom-test.txt', ]))
    datafiles.append((dirprefix+'asciidoc3/tests/docbook_validation',
                      ['tests/docbook_validation/asciidoc3_docbook45_validation.py',
                       'tests/docbook_validation/asciidoc3_docbook51_relaxng_validation.py',
                       'tests/docbook_validation/asciidoc3_docbook51_schematron_validation.py',
                       'tests/docbook_validation/asciidoc3_docbook51_w3cxml_validation.py',
                       'tests/docbook_validation/rng-docbook51.rng', ]))

with open("README.md", "r") as fh:
    long_description = fh.read()

setup(
    name="asciidoc3",
    version="3.2.0",
    description="""AsciiDoc3 Python3 GNU/Linux Windows AsciiDoc - see https://asciidoc3.org/pypi.html BEFORE installing""",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords=['text', 'markup', 'Windows', 'asciidoc', 'asciidoc3', 'python3'],
    author="Berthold Gehrke",
    author_email="berthold.gehrke@gmail.com",
    url="https://asciidoc3.org",
    project_urls={
        "Source": "https://gitlab.com/asciidoc3/asciidoc3",
        "Funding": "https://asciidoc3.org/contact.html"
    },
    license='GPLv2+',
    packages=find_packages(),
    entry_points={'console_scripts':
                  ['asciidoc3=asciidoc3.asciidoc3:main',
                   'a2x3=asciidoc3.a2x3:main',
                   'asciidoc3_postinstall=asciidoc3.asciidoc3_postinstall:main']},
    include_package_data=True,
    # 'data_files' do not contain the symlinks to dir 'images' and to 'asciidoc3api.py'
    # from ./tests: so run 'asciidoc3_postinstall' after 'pip install --user asciidoc3'
    data_files=datafiles,
    zip_safe=False,
    python_requires=">= 3.5",
    classifiers=[  # for a list of valid classifiers see https://pypi.org/classifiers/
        'Topic :: Text Processing',
        'Topic :: Text Processing :: Markup',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Natural Language :: English',
        'Operating System :: POSIX',
        'Operating System :: Microsoft :: Windows',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Development Status :: 5 - Production/Stable',
    ],
)



#!/usr/bin/env python3

'''
./tests/docbook_validation/asciidoc3_docbook45_validation.py
validation against DTD DocBook4.5

Tests all DocBook45-files produced with testasciidoc3.py found
in directory './tests/data' if valid against the
DocBook45 DTD.
We are using the following tools to validate:
- xmllint     (Package libxml2-utils)
- xmlstarlet  (Package xmlstarlet)
- lxml.de     (Package python3-lxml)
Check if present on command line with 'xmllint' and 'xmlstarlet' and
with 'import lxml'.
You can disable any program that is not available, but it is recommended
to make use of at least two validators to make the test more reliable.

USAGE:
'python3 asciidoc3_docbook45_validation.py'
to test all files ending with 'docbook45.xml' in directory './tests/data' or
'python3 asciidoc3_docbook45_validation.py path/to/db45_xml_file.xml'
to validate just one file.

Copyright: (c) 2020 Berthold Gehrke <berthold.gehrke@gmail.com>
License:   GNU GPL v2 or higher
'''

import os
from pathlib import Path
import subprocess
import sys
try:
    from lxml import etree
    LXMLPARSER = 'lxmlparser'
except ModuleNotFoundError:
    LXMLPARSER = ''

# External executables.
XMLLINT = 'xmllint'         # Set to '' to disable.
# XMLLINT = ''
XMLSTARLET = 'xmlstarlet'   # Set to '' to disable.
# XMLSTARLET = ''
# LXMLPARSER = ''  # Uncomment to disable lxmlparser.


def xmllint_db45(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    subprocess.run(["xmllint", "--nonet", "--noout", "--valid",
                    path_to_data + testfile])


def xmlstarlet_db45(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    subprocess.run(["xmlstarlet", "val", "--dtd",
                    "/usr/share/xml/docbook/schema/dtd/4.5/docbookx.dtd",
                    "-q", path_to_data + testfile])


def lxml_db45(testfile):
    '''
    xmmlint uses xml catalog.
    Be sure to have package 'docbook' installed.
    '''
    db45_parser = etree.XMLParser(dtd_validation=True)
    try:
        etree.parse(path_to_data + testfile, db45_parser)
    except etree.XMLSyntaxError:
        print("lxml reports", testfile, "******* invalid *******")


def main():
    if filepath:
        if XMLLINT:
            xmllint_db45(base_filepath)
        if XMLSTARLET:
            xmlstarlet_db45(base_filepath)
        if LXMLPARSER:
            lxml_db45(base_filepath)
    else:
        with os.scandir(path_to_data) as testpath:
            for entry in testpath:
                if entry.name.endswith('docbook45.xml'):
                    if XMLLINT:
                        xmllint_db45(entry.name)
                    if XMLSTARLET:
                        xmlstarlet_db45(entry.name)
                    if LXMLPARSER:
                        lxml_db45(entry.name)


if __name__ == '__main__':
    filepath = ''
    full_path = os.path.abspath(os.path.dirname(__file__))
    if len(sys.argv) == 2:
        filepath = sys.argv[1]
        if filepath.endswith('/'):
            sys.exit("FATAL: Please enter path without trailing '/'!")
        base_filepath = os.path.basename(filepath)
        dirname = os.path.dirname(filepath)
        if dirname:
            path_to_data = dirname + "/"
        else:
            path_to_data = ""
    elif len(sys.argv) > 2:
        sys.exit("FATAL: Only one argument (path) possible!")
    elif len(sys.argv) == 1:
        path_to_tests = str(Path(full_path).parents[0])
        path_to_data = path_to_tests + "/data/"
    else:
        sys.exit("FATAL: Found no argument at all!")
    main()

